import asyncio
import dataclasses
import pathlib
import re
import tempfile
import typing

import cookiecutter.main
import flatdict
import sprockets_postgres
from cookiecutter import exceptions
from jinja2 import exceptions as jinja_exceptions

from imbi import models, oauth2
from imbi.automations import base
from imbi.clients import gitlab, sonarqube
if typing.TYPE_CHECKING:
    from imbi import app, user


class GitLabCreateProjectAutomation(base.Automation):

    INSERT_SQL = re.sub(r'\s+', ' ', """\
        INSERT INTO v1.project_links
                    (project_id, link_type_id, created_by, url)
             VALUES (%(project_id)s, %(link_type_id)s,
                     %(username)s, %(url)s)""")

    UPDATE_SQL = re.sub(r'\s+', ' ', """\
        UPDATE v1.projects
           SET gitlab_project_id = %(gitlab_project_id)s
         WHERE id = %(project_id)s""")

    def __init__(self,
                 application: 'app.Application',
                 project_id: int,
                 current_user: 'user.User',
                 db: sprockets_postgres.PostgresConnector):
        super().__init__(application, current_user, db)
        self.imbi_project_id = project_id
        self._gitlab: typing.Optional[gitlab.GitLabClient] = None
        self._gitlab_parent: typing.Optional[dict] = None
        self._project: typing.Optional[models.Project] = None

    async def prepare(self) -> typing.List[str]:
        project: models.Project
        token: oauth2.IntegrationToken
        project, token = await asyncio.gather(
            self._get_project(self.imbi_project_id),
            self._get_gitlab_token())
        if project is not None and project.gitlab_project_id is not None:
            self._add_error(
                'GitLab project {} already exists for {}',
                project.gitlab_project_id, project.slug)
        elif project is not None and token is not None:
            self._gitlab = gitlab.GitLabClient(token, self.application)
            self._gitlab_parent = await self._get_gitlab_parent(project)
            self._project = project
        return self.errors

    async def run(self) -> gitlab.ProjectInfo:
        project = await self._gitlab.create_project(
            self._gitlab_parent,
            self._project.name,
            description=self._project.description)
        await self.db.execute(
            self.UPDATE_SQL,
            {
                'project_id': self._project.id,
                'gitlab_project_id': project.id,
            })

        link_id = self.automation_settings['gitlab'].get(
            'project_link_type_id')
        if link_id:
            await self.db.execute(
                self.INSERT_SQL,
                {
                    'link_type_id': link_id,
                    'project_id': self._project.id,
                    'url': project.web_url,
                    'username': self.user.username,
                })
        return project

    async def _get_project(
            self, project_id: int) -> typing.Optional[models.Project]:
        project = await super()._get_project(project_id)
        if project is not None:
            if project.namespace.gitlab_group_name is None:
                self._add_error('missing GitLab group for namespace {}',
                                project.namespace.slug)
            if project.project_type.gitlab_project_prefix is None:
                self._add_error('missing no GitLab prefix for project type {}',
                                project.project_type.slug)
        return None if self._has_error() else project

    async def _get_gitlab_parent(self, project:  models.Project) -> dict:
        gitlab_parent = await self._gitlab.fetch_group(
            project.namespace.gitlab_group_name,
            project.project_type.gitlab_project_prefix)
        if not gitlab_parent:
            self._add_error(
                'GitLab path {}/{} does not exist',
                project.namespace.gitlab_group_name,
                project.project_type.gitlab_project_prefix)
        return gitlab_parent


class CookieCutterError(Exception):
    """Raised when there is an error applying the cookiecutter"""


class InitialCommitError(Exception):
    """Raised when there is an error creating the initial commit"""


class GitLabInitialCommitAutomation(base.Automation):

    GET_COOKIE_CUTTER = re.sub(r'\s+', ' ', """\
        SELECT name, project_type_id, url
          FROM v1.cookie_cutters
         WHERE type='project'
           AND url = %(url)s""")

    def __init__(self,
                 application: 'app.Application',
                 project_id: int,
                 cookie_cutter: str,
                 current_user: 'user.User',
                 db: sprockets_postgres.PostgresConnector):
        super().__init__(application, current_user, db)
        self._cookie_cutter_name = cookie_cutter
        self._imbi_project_id = project_id
        self._cookie_cutter: typing.Optional[models.CookieCutter] = None
        self._gitlab: typing.Optional[gitlab.GitLabClient] = None
        self._gitlab_project: typing.Optional[gitlab.ProjectInfo] = None
        self._project: typing.Optional[models.Project] = None
        self._token: typing.Optional[oauth2.IntegrationToken] = None

    async def prepare(self) -> typing.List[str]:
        self._project = await self._get_project(self._imbi_project_id)
        self._token = await self._get_gitlab_token()
        self._cookie_cutter = await self._get_cookie_cutter(
            self._cookie_cutter_name)

        if not self._has_error() \
                and (self._project.type.id
                     != self._cookie_cutter.project_type_id):
            self._add_error(
                'Cookie Cutter {} is not available for Project Type {}',
                self._cookie_cutter.name, self._project.type.slug)
        elif self._project.gitlab_project_id is None:
            self._add_error('GitLab project does not exist for {}',
                            self._project.slug)
        else:
            self._gitlab = gitlab.GitLabClient(self._token, self.application)
            self._gitlab_project = await self._gitlab.fetch_project(
                self._project.gitlab_project_id)
            if self._gitlab_project is None:
                self._add_error('GitLab project id {} does not exist',
                                self._project.gitlab_project_id)
        return self.errors

    async def run(self):
        self.logger.info('generating initial commit for %s (%s) from %s',
                         self._project.slug, self._project.id,
                         self._cookie_cutter.url)
        context = dataclasses.asdict(self._project)
        links = {k.lower().replace(' ', '_'): v
                 for k, v in context['links'].items()}
        urls = {k.lower().replace(' ', '_'): v
                for k, v in context['urls'].items()}
        context.update({
            'environments': ','.join(self._project.environments),
            'gitlab': {
                'namespace_id': self._gitlab_project.namespace.id,
                'project_id': self._project.gitlab_project_id
            },
            'links': links,
            'pagerduty': {
                'service_id': None,
            },
            'sentry': {
                'dashboard': None,
                'dsn': None,
                'organization': None,
                'slug': self._project.sentry_project_slug,
                'team': None
            },
            'sonarqube': {'key': None},
            'urls': urls
        })
        if self.automation_settings['sonarqube'].get('url'):
            context.update({
                'sonarqube': {'key': sonarqube.generate_key(self._project)}
            })

        for var in ['gitlab_project_id', 'pagerduty_service_id',
                    'sentry_project_slug', 'sonarqube_project_key']:
            del context[var]

        self.logger.debug('Context %r', {'project': context})
        with tempfile.TemporaryDirectory() as tmp_dir:
            self.logger.debug('expanding %s for project %s in %s',
                              self._cookie_cutter.url, self._project.id,
                              tmp_dir)
            try:
                project_dir = cookiecutter.main.cookiecutter(
                    self._cookie_cutter.url,
                    extra_context=dict(
                        flatdict.FlatDict({'project': context}, '_')),
                    no_input=True,
                    output_dir=tmp_dir)
            except (exceptions.ContextDecodingException,
                    exceptions.NonTemplatedInputDirException,
                    exceptions.UndefinedVariableInTemplate,
                    jinja_exceptions.UndefinedError) as error:
                raise CookieCutterError(str(error))

            project_dir = pathlib.Path(project_dir)

            """Disabling for the time being
            self.logger.debug('reformatting project files')
            isort_cfg = self.automation_settings['isort']
            isort_cfg.setdefault('known_first_party', [])
            isort_cfg['known_first_party'].append(context['package_name'])
            yapf_style = self.automation_settings['yapf']

            for py_file in project_dir.rglob('*.py'):
                isort.api.sort_file(py_file, **isort_cfg)
                yapf_api.FormatFile(
                    str(py_file), style_config=yapf_style,
                    in_place=True, logger=None)
            """

            self.logger.debug('committing to GitLab')
            commit_info = await self._gitlab.commit_tree(
                self._gitlab_project, project_dir,
                'Initial commit (automated)')

            return commit_info

    async def _get_cookie_cutter(self, url: str) \
            -> typing.Optional[models.CookieCutter]:
        result = await self.db.execute(
            self.GET_COOKIE_CUTTER, {'url': url})
        if result.row_count != 0:
            return models.CookieCutter(**result.row)
        else:
            self._add_error('Cookie cutter {} does not exist', url)



import logging
import typing

import sprockets_postgres

from imbi import models, oauth2
if typing.TYPE_CHECKING:
    from imbi import app, user


class Automation:

    def __init__(self,
                 application: 'app.Application',
                 current_user: 'user.User',
                 db: sprockets_postgres.PostgresConnector):
        self.application = application
        self.automation_settings = self.application.settings['automations']
        self.db = db
        self.errors: typing.List[str] = []
        self.logger = logging.getLogger(__package__).getChild(
            self.__class__.__name__)
        self.user = current_user

    def _add_error(self, msg_format, *args):
        message = msg_format.format(*args)
        self.logger.warning('%s', message)
        self.errors.append(message)

    def _has_error(self) -> bool:
        return len(self.errors) > 0

    async def _get_project(self, project_id: int) \
            -> typing.Optional[models.Project]:
        project = await models.project(project_id, self.application)
        if not project:
            self._add_error('project not found for {}', project_id)
        return None if self._has_error() else project

    async def _get_gitlab_token(self) \
            -> typing.Optional[oauth2.IntegrationToken]:
        tokens = await self.user.fetch_integration_tokens('gitlab')
        if not tokens:
            self._add_error('GitLab token not found for current user')
            return None
        return tokens[0]



"""
User Model supporting both LDAP and PostgreSQL data sources

"""
import dataclasses
import datetime
import logging
import re
import typing
from itertools import chain

import ldap3
import psycopg2.errors
from tornado import web

from imbi import ldap, oauth2, timestamp

LOGGER = logging.getLogger(__name__)


class Group:
    """Group class to represent a single group a user is a member of"""
    __slots__ = ['name', 'permissions']

    def __init__(self, name: str, permissions: typing.List[str]):
        self.name = name
        self.permissions = sorted(permissions or [])

    def __iter__(self):
        return iter([('name', self.name), ('permissions', self.permissions)])

    def __repr__(self):
        return '<Group name={} permissions={}>'.format(
            self.name, self.permissions)


@dataclasses.dataclass
class ConnectedIntegration:
    """The connection between a user and a specific integration."""
    name: str
    external_id: str


class User:
    """Holds the user attributes and interfaces with the directory server"""

    REFRESH_AFTER = datetime.timedelta(minutes=5)

    SQL_AUTHENTICATE = re.sub(r'\s+', ' ', """\
           UPDATE v1.users
              SET last_seen_at = CURRENT_TIMESTAMP
            WHERE username = %(username)s
              AND password = %(password)s
              AND user_type = 'internal'
        RETURNING username""")

    SQL_AUTHENTICATE_TOKEN = re.sub(r'\s+', ' ', """\
           UPDATE v1.users
              SET last_seen_at = CURRENT_TIMESTAMP
            WHERE username IN (
                    SELECT username
                      FROM v1.authentication_tokens
                     WHERE token = %(token)s
                       AND expires_at > CURRENT_TIMESTAMP)
        RETURNING username, user_type, external_id""")  # nosec

    SQL_GROUPS = re.sub(r'\s+', ' ', """\
        SELECT a.name, a.permissions
          FROM v1.groups AS a
          JOIN v1.group_members AS b ON b.group = a.name
         WHERE b.username = %(username)s""")

    SQL_INTEGRATIONS = re.sub(r'\s+', ' ', """\
        SELECT integration AS name, external_id
          FROM v1.user_oauth2_tokens
         WHERE username = %(username)s""")

    SQL_REFRESH = re.sub(r'\s+', ' ', """\
        SELECT username, created_at, last_seen_at, user_type, external_id,
               email_address, display_name
          FROM v1.users
         WHERE username = %(username)s""")

    SQL_UPDATE_GROUP_MEMBERSHIPS_FROM_LDAP = re.sub(r'\s+', ' ', """\
        SELECT maintain_group_membership_from_ldap_groups AS groups
          FROM maintain_group_membership_from_ldap_groups(%(username)s,
                                                          %(groups)s)""")

    SQL_UPDATE_LAST_SEEN_AT = re.sub(r'\s+', ' ', """\
            UPDATE v1.users
               SET last_seen_at = CURRENT_TIMESTAMP
             WHERE username = %(username)s
         RETURNING last_seen_at""")

    SQL_UPDATE_USER_FROM_LDAP = re.sub(r'\s+', ' ', """\
        INSERT INTO v1.users (username, user_type, external_id,
                              display_name, email_address, last_seen_at)
             VALUES (%(username)s,
                     %(user_type)s,
                     %(external_id)s,
                     %(display_name)s,
                     %(email_address)s,
                     CURRENT_TIMESTAMP)
        ON CONFLICT (username)
                 DO UPDATE SET email_address = EXCLUDED.email_address,
                                 external_id = EXCLUDED.external_id,
                                   user_type = EXCLUDED.user_type,
                                    password = NULL,
                                last_seen_at = CURRENT_TIMESTAMP
                        WHERE users.username = EXCLUDED.username
          RETURNING last_seen_at""")

    def __init__(self, application,
                 username: typing.Optional[str] = None,
                 password: typing.Optional[str] = None,
                 token: typing.Optional[str] = None) -> None:
        self._application = application
        self._ldap = ldap.Client(application.settings['ldap'])
        self._ldap_conn: typing.Optional[ldap3.Connection] = None
        self.username = username
        self.created_at: typing.Optional[datetime.datetime] = None
        self.last_refreshed_at: typing.Optional[datetime.datetime] = None
        self.last_seen_at: typing.Optional[datetime.datetime] = None
        self.user_type = 'internal'
        self.external_id: typing.Optional[str] = None
        self.email_address: typing.Optional[str] = None
        self.display_name: typing.Optional[str] = None
        self._password = password
        self.token = token
        self.connected_integrations: typing.List[ConnectedIntegration] = []
        self.groups: typing.List[Group] = []
        self.permissions: typing.List[str] = []

    @property
    def password(self):
        return self._password

    def __del__(self):
        if self._ldap_conn is not None:
            self._ldap_conn.strategy.close()
            self._ldap_conn = None

    def __repr__(self):
        return '<User username={} user_type={} permissions={}>'.format(
            self.username, self.user_type, self.permissions)

    def as_dict(self) -> dict:
        """Return a representation of the user data as a dict"""
        return {
            'username': self.username,
            'user_type': self.user_type,
            'external_id': self.external_id,
            'created_at': timestamp.isoformat(self.created_at),
            'last_seen_at': timestamp.isoformat(self.last_seen_at),
            'email_address': self.email_address,
            'display_name': self.display_name,
            'groups': [g.name for g in self.groups],
            'password': self._application.encrypt_value(
                '' if self.password is None else self.password),
            'permissions': self.permissions,
            'last_refreshed_at': timestamp.isoformat(self.last_refreshed_at),
            'integrations': sorted({
                app.name for app in self.connected_integrations}),
        }

    async def authenticate(self) -> bool:
        """Validate that the current session is valid. Returns boolean
        if successful.

        """
        if self.token:
            return await self._token_auth()
        result = await self._db_auth()
        if not result and self._ldap.is_enabled:
            return await self._ldap_auth()
        return result

    def has_permission(self, value: str) -> bool:
        """Check all of the permissions assigned to the user, returning
        `True` if the role is assigned through any group memberships

        :param value: The role to check for

        """
        return value in self.permissions

    @staticmethod
    def on_postgres_error(_metric_name: str, exc: Exception) -> None:
        raise web.HTTPError(500, 'System error')

    @classmethod
    def on_token_postgres_error(cls, metric_name: str, exc: Exception) -> None:
        """Pass InvalidTextRepresentation exceptions through."""
        if isinstance(exc, psycopg2.errors.InvalidTextRepresentation):
            raise exc
        return cls.on_postgres_error(metric_name, exc)

    async def refresh(self) -> None:
        """Update the attributes from LDAP"""
        if self.external_id and self._ldap.is_enabled:
            await self._ldap_refresh()
        else:
            await self._db_refresh()

    @property
    def should_refresh(self) -> bool:
        """Returns True if the amount of time that has passed since the last
        refresh has exceeded the threshold.

        """
        age = timestamp.age(self.last_refreshed_at)
        LOGGER.debug('Last refresh: %s < %s == %s, %s < %s == %s',
                     self.REFRESH_AFTER, age, self.REFRESH_AFTER < age,
                     self.last_refreshed_at, self.last_seen_at,
                     self.last_refreshed_at < self.last_seen_at)
        return ((self.REFRESH_AFTER < age) or
                (self.last_refreshed_at < self.last_seen_at))

    async def update_last_seen_at(self) -> None:
        """Update the last_seen_at column in the database for the user"""
        async with self._application.postgres_connector(
                on_error=self.on_postgres_error) as conn:
            result = await conn.execute(
                self.SQL_UPDATE_LAST_SEEN_AT,
                {'username': self.username},
                'user-update-last-seen-at')
            self.last_seen_at = result.row['last_seen_at']

    async def fetch_integration_tokens(self, integration: str) \
            -> typing.List[oauth2.IntegrationToken]:
        """Retrieve access tokens for the specified integration."""
        obj = await oauth2.OAuth2Integration.by_name(
            self._application, integration)
        return await obj.get_user_tokens(self)

    def _assign_values(self, data: dict) -> None:
        """Assign values from the cursor row to the instance"""
        for key, value in data.items():
            setattr(self, key, value)

    async def _db_auth(self) -> bool:
        """Validate via the v1.users table in the database"""
        LOGGER.debug('Authenticating with the database')
        async with self._application.postgres_connector(
                on_error=self.on_postgres_error) as conn:
            password = None
            if self.password is not None:
                password = self._application.hash_password(self.password)
            result = await conn.execute(
                self.SQL_AUTHENTICATE,
                {'username': self.username, 'password': password},
                'user-authenticate')
            if not result.row_count:
                self._reset()
                return False
        await self._db_refresh()
        return True

    async def _db_groups(self) -> typing.List[Group]:
        """Return the groups for the user as a list of group objects"""
        async with self._application.postgres_connector(
                on_error=self.on_postgres_error) as conn:
            result = await conn.execute(
                self.SQL_GROUPS, {'username': self.username},
                'user-groups')
            return [Group(**r) for r in result]

    async def _db_refresh(self) -> None:
        """Fetch the latest values from the database"""
        async with self._application.postgres_connector(
                on_error=self.on_postgres_error) as conn:
            result = await conn.execute(
                self.SQL_REFRESH, {'username': self.username},
                'user-refresh')
        if result:
            self._assign_values(result.row)
            self.groups = await self._db_groups()
            self.permissions = sorted(set(
                chain.from_iterable([g.permissions for g in self.groups])))
            self.connected_integrations = await self._refresh_integrations()
            self.last_refreshed_at = max(
                timestamp.utcnow(), self.last_seen_at or timestamp.utcnow())
        else:
            self._reset()

    async def _ldap_auth(self) -> bool:
        """Authenticate via LDAP"""
        if not self.password:
            LOGGER.debug('Can not LDAP authenticate without a password')
            return False
        LOGGER.debug('Authenticating via LDAP')
        if not self._ldap_conn:
            self._ldap_conn = await self._ldap.connect(self.username,
                                                       self.password)
        if self._ldap_conn:
            LOGGER.debug('Authenticated as %s', self.username)
            self.user_type = 'ldap'
            result = self._ldap_conn.extend.standard.who_am_i()
            self.external_id = result[3:].strip()
            await self._ldap_refresh()
            return True
        self._reset()
        return False

    async def _ldap_refresh(self) -> None:
        """Update the attributes from LDAP"""
        LOGGER.debug('Refreshing attributes from LDAP server')
        attrs = await self._ldap.attributes(self._ldap_conn, self.external_id)
        LOGGER.debug('Attributes: %r', attrs)
        self.display_name = attrs.get('displayName', attrs['cn'])
        self.email_address = attrs.get('mail')

        async with self._application.postgres_connector(
                on_error=self.on_postgres_error) as conn:
            result = await conn.execute(
                self.SQL_UPDATE_USER_FROM_LDAP,
                {
                    'username': self.username,
                    'user_type': 'ldap',
                    'external_id': self.external_id,
                    'display_name': self.display_name,
                    'email_address': self.email_address
                }, 'user-update')
            self.last_seen_at = result.row['last_seen_at']

        # Update the groups in the database and get the group names
        ldap_groups = await self._ldap.groups(self._ldap_conn,
                                              self.external_id)
        async with self._application.postgres_connector(
                on_error=self.on_postgres_error) as conn:
            await conn.callproc(
                'v1.maintain_group_membership_from_ldap_groups',
                (self.username, ldap_groups),
                'user-maintain-groups')

        # Update the groups attribute
        self.groups = await self._db_groups()
        self.permissions = sorted(set(
            chain.from_iterable([g.permissions for g in self.groups])))
        self.connected_integrations = await self._refresh_integrations()
        self.last_refreshed_at = max(timestamp.utcnow(), self.last_seen_at)

    async def _refresh_integrations(self) -> typing.Sequence[
            ConnectedIntegration]:
        """Fetch connected integration details from the DB."""
        async with self._application.postgres_connector(
                on_error=self.on_postgres_error) as conn:
            result = await conn.execute(
                self.SQL_INTEGRATIONS, {'username': self.username})
            return [ConnectedIntegration(**r) for r in result]

    def _reset(self) -> None:
        """Reset the internally assigned values associated with this user
        object.

        """
        self._ldap_conn = None
        self.created_at = None
        self.last_seen_at = None
        self.user_type = 'internal'
        self.external_id = None
        self.email_address = None
        self.display_name = None
        self.groups = []
        self.permissions = []
        self.last_refreshed_at = None
        self.connected_integrations = []

    async def _token_auth(self) -> bool:
        """Validate via v1.authentication_tokens table"""
        LOGGER.debug('Authenticating with token: %s', self.token)
        async with self._application.postgres_connector(
                on_error=self.on_token_postgres_error) as conn:
            try:
                result = await conn.execute(
                    self.SQL_AUTHENTICATE_TOKEN, {'token': self.token},
                    'user-token-auth')
            except psycopg2.errors.InvalidTextRepresentation:
                return False
            if not result.row_count:
                return False
            self._assign_values(result.row)
        await self._db_refresh()
        return True



import re

from imbi.endpoints import base


class CollectionRequestHandler(base.CollectionRequestHandler):

    NAME = 'project-activity-feed'
    ID = 'project_id'
    COLLECTION_SQL = re.sub(r'\s+', ' ', """\
        WITH created AS (
                SELECT created_at AS "when",
                       created_by AS who,
                       'created' AS what,
                       NULL AS fact_name,
                       NULL AS value
                  FROM v1.projects
                 WHERE id = %(project_id)s),
             updated AS (
                SELECT last_modified_at AS "when",
                       last_modified_by AS who,
                       'updated' AS what,
                       NULL AS fact_name,
                       NULL AS value
                  FROM v1.projects
                 WHERE id = %(project_id)s),
             facts AS (
                SELECT a.recorded_at AS "when",
                       a.recorded_by AS who,
                       'updated fact' AS what,
                       b.name AS fact_name,
                       a.value
                  FROM v1.project_facts AS a
                  JOIN v1.project_fact_types AS b
                    ON b.id = a.fact_type_id
                 WHERE a.project_id = %(project_id)s),
            combined AS (
                SELECT *
                  FROM created
                 UNION
                SELECT *
                  FROM updated
                 UNION
                SELECT *
                  FROM facts)
        SELECT a.when,
               a.who,
               b.display_name,
               a.what,
               a.fact_name,
               a.value
          FROM combined AS a
          JOIN v1.users AS b
            ON b.username = a.who
         ORDER BY a.when DESC;""")



"""
Authenticated User Session

"""
import json
import logging
import typing
import uuid

import aioredis
from sprockets.http import app
from tornado import web

from imbi import timestamp, user

LOGGER = logging.getLogger(__name__)

DEFAULT_POOL_SIZE = 5


class Session:
    """Session object manages session state and the user object."""

    def __init__(self, handler: web.RequestHandler) -> None:
        self._handler = handler
        self.authenticated = False
        self.id = self._get_id_from_cookie() or str(uuid.uuid4())
        self.last_save = None
        self.start = timestamp.isoformat()
        self.user = None

    async def authenticate(self, username: str, password: str) -> bool:
        """Authenticate the user and attach it to the session

        :param username: The username to authenticate with
        :param password: The password to use when authenticating

        """
        self.user = user.User(
            self._handler.application, username=username, password=password)
        self.authenticated = await self.user.authenticate()
        if not self.authenticated:
            self.user = None
            await self.clear()
            return False
        return True

    async def initialize(self) -> None:
        """Used to initialize the data, loading the data if it is set"""
        self.user = await self._load_data()
        if not self.user:
            LOGGER.debug('Session initialized without a user')
            return

        self.authenticated = await self.user.authenticate()
        if not self.authenticated:
            self.user = None
            return

        if self.user.should_refresh:
            await self.user.refresh()
            await self.save()

    async def clear(self) -> None:
        """Clear out the currently loaded session by clearing the cookie
        and removing the data from redis.

        """
        LOGGER.debug('Deleting session %s', self.id)
        self._handler.clear_cookie('session')
        await self._redis.delete(self._redis_key)
        self.authenticated = False
        self.last_save = None
        self.start = timestamp.isoformat()
        self.user = None

    async def save(self) -> None:
        """Save session data to redis"""
        LOGGER.debug('Saving session %s', self.id)
        user_data = {} if not self.user else self.user.as_dict()
        await self._redis.set(
            self._redis_key,
            json.dumps({
                'user': user_data,
                'last_save': timestamp.isoformat(),
                'start': self.start}),
            expire=self._settings['session_duration'] * 86400)
        self._handler.set_secure_cookie(
            'session', self.id,
            expires_days=self._settings['session_duration'])

    @property
    def _application(self) -> app.Application:
        """Return the application  instance"""
        return self._handler.application

    def _get_id_from_cookie(self) -> typing.Optional[str]:
        """Get the Session ID from the secure cookie, decoding it from
        UTF-8, if set.

        """
        value = self._handler.get_secure_cookie('session')
        if value:
            return value.decode('utf-8')

    async def _load_data(self) -> typing.Optional['user.User']:
        """Load the data from Redis, creating the user object and returning it
        if there was a previously saved user,

        :rtype: User or None
        """
        LOGGER.debug('Loading session %s', self.id)
        result = await self._redis.get(self._redis_key)
        if not result:
            LOGGER.info('Session %r not found', self.id)
            return
        data = json.loads(result.decode('utf-8'))
        self.last_save = data['last_save']
        self.start = data['start']
        if not data.get('user'):
            return

        password = data['user'].pop('password', None)
        if password is not None:
            password = self._application.decrypt_value(password)

        user_obj = user.User(self._handler.application, password=password)
        for key, value in data['user'].items():
            setattr(user_obj, key, value)
        return user_obj

    @property
    def _redis(self) -> aioredis.Redis:
        """Return the handle to the Redis client"""
        return self._application.session_redis

    @property
    def _redis_key(self) -> str:
        """Return the properly formatted session key."""
        return 'session:{}'.format(self.id)

    @property
    def _settings(self) -> dict:
        """Return the application settings dict"""
        return self._application.settings



import base64

from cryptography import fernet
from cryptography.hazmat.primitives import hashes, hmac


class DecryptionFailure(RuntimeError):
    """Raised when a value cannot be decrypted."""
    ...


class Keychain:
    """Simple keychain abstraction.

    This class takes a random 32-byte key and uses it for symmetric
    encryption and one-way hashing.  The interfaces are ALWAYS in
    terms of byte strings.

    .. rubric:: Import Note

    Though this keychain can decrypt any message that was encrypted
    using the same key, it is not true that encrypting the same plain
    text twice will result in the same cipher text value.  In fact,
    the current implementation will guarantee that this is not the
    case so you should never store an encrypted value somewhere and
    then compare it to a newly encrypted value for raw byte equality.

    """
    algorithm = hashes.SHA512()

    def __init__(self, key: bytes):
        if len(key) != 32:
            raise ValueError('Keychain requires a 32-byte key')
        self._key = key
        self._fernet = fernet.Fernet(base64.urlsafe_b64encode(key))

    def encrypt(self, plaintext: bytes) -> bytes:
        """Encrypt `ciphertext` using the configured key."""
        return base64.urlsafe_b64decode(self._fernet.encrypt(plaintext))

    def decrypt(self, ciphertext: bytes) -> bytes:
        """Decrypt `ciphertext` using the configured key.

        :raises: :exc:`DecryptionFailure` when the value cannot be
            decrypted.
        """
        try:
            return self._fernet.decrypt(base64.urlsafe_b64encode(ciphertext))
        except fernet.InvalidToken as error:
            raise DecryptionFailure() from error

    def hash(self, data: str) -> bytes:
        """Generate a HMAC-SHA512 of the supplied data."""
        hasher = hmac.HMAC(self._key, self.algorithm)
        hasher.update(data.encode('utf-8'))
        return hasher.finalize()



"""
Integration Endpoints

"""
from tornado import web

from . import gitlab, oauth2

URLS = [
    web.url(r'^/gitlab/auth', gitlab.RedirectHandler),
    web.url(r'^/gitlab/namespaces', gitlab.UserNamespacesHandler),
    web.url(r'^/gitlab/projects', gitlab.ProjectsHandler),
    web.url(r'^/integrations$', oauth2.CollectionRequestHandler),
    web.url(r'^/integrations/(?P<name>[\w_\-%\+]+)$',
            oauth2.RecordRequestHandler, name='integration')
]



from imbi.automations import sentry
from imbi.endpoints import base
from imbi.endpoints.ui.automations import mixins
from imbi.opensearch import project


class ProjectCreationRequestHandler(mixins.PrepareFailureMixin,
                                    project.RequestHandlerMixin,
                                    base.ValidatingRequestHandler):
    NAME = 'ui-sentry-create-project'

    async def post(self) -> None:
        request = self.get_request_body()
        project_id = int(request['project_id'])
        user = await self.get_current_user()  # never None
        async with self.postgres_transaction() as transaction:
            client = sentry.SentryCreateProjectAutomation(
                self.application, project_id, user, transaction)
            failures = await client.prepare()
            if failures:
                raise self.handle_prepare_failures('Create Project', failures)

            project_info = await client.run()
            self.send_response({'sentry_project_url': project_info.link})

        await self.index_document(project_id)



import re

from imbi.endpoints import base


class CollectionRequestHandler(base.CollectionRequestHandler):

    NAME = 'integrations'
    ITEM_NAME = 'integration'
    ID_KEY = 'name'
    FIELDS = ['name', 'api_endpoint', 'callback_url', 'authirization_endpoint',
              'token_endpoint', 'revoke_endpoint', 'client_id',
              'client_secret', 'public_client']

    COLLECTION_SQL = re.sub(r'\s+', ' ', """\
        SELECT name, api_endpoint, callback_url, authorization_endpoint,
               token_endpoint, revoke_endpoint, client_id
          FROM v1.oauth2_integrations
         ORDER BY "name" ASC""")

    POST_SQL = re.sub(r'\s+', ' ', """\
        INSERT INTO v1.oauth2_integrations
                    (name, api_endpoint, callback_url, authorization_endpoint,
                     token_endpoint, revoke_endpoint, client_id, client_secret,
                     public_client)
             VALUES (%(name)s, %(api_endpoint)s, %(callback_url)s,
                     %(authorization_endpoint)s, %(token_endpoint)s,
                     %(revoke_endpoint)s, %(client_id)s, %(client_secret)s,
                     %(public_client)s)
          RETURNING name""")

    GET_SQL = re.sub(r'\s+', ' ', """\
        SELECT name, api_endpoint, callback_url, authorization_endpoint,
               token_endpoint, revoke_endpoint, client_id
          FROM v1.oauth2_integrations
         WHERE name = %(name)s""")


class RecordRequestHandler(base.CRUDRequestHandler):

    NAME = 'integration'
    ID_KEY = 'name'

    GET_SQL = re.sub(r'\s+', ' ', """\
        SELECT name, api_endpoint, callback_url, authorization_endpoint,
               token_endpoint, revoke_endpoint, client_id
          FROM v1.oauth2_integrations
         WHERE name = %(name)s""")

    DELETE_SQL = re.sub(r'\s+', ' ', """\
        DELETE FROM v1.oauth2_integrations WHERE name = %(name)s""")



import re

from imbi.endpoints import base


class _RequestHandlerMixin:

    ID_KEY = ['id']
    FIELDS = ['id', 'project_type_ids', 'name', 'fact_type', 'data_type',
              'description', 'ui_options', 'weight']
    DEFAULTS = {
        'data_type': 'string',
        'fact_type': 'free-form',
        'weight': 0
    }

    GET_SQL = re.sub(r'\s+', ' ', """\
        SELECT id, created_at, created_by, last_modified_at, last_modified_by,
               project_type_ids, name, fact_type, data_type, description,
               ui_options, weight
          FROM v1.project_fact_types
         WHERE id=%(id)s""")


class CollectionRequestHandler(_RequestHandlerMixin,
                               base.CollectionRequestHandler):

    NAME = 'fact-types'
    ITEM_NAME = 'fact-type'

    COLLECTION_SQL = re.sub(r'\s+', ' ', """\
        SELECT id, project_type_ids, name, fact_type, data_type, description,
               ui_options, weight
          FROM v1.project_fact_types
         ORDER BY name, project_type_ids""")

    POST_SQL = re.sub(r'\s+', ' ', """\
        INSERT INTO v1.project_fact_types
                    (project_type_ids, created_by, name, fact_type, data_type,
                     description, ui_options, weight)
             VALUES (%(project_type_ids)s, %(username)s, %(name)s,
                     %(fact_type)s, %(data_type)s, %(description)s,
                     %(ui_options)s, %(weight)s)
          RETURNING id""")


class RecordRequestHandler(_RequestHandlerMixin, base.AdminCRUDRequestHandler):

    NAME = 'fact-type'

    DELETE_SQL = 'DELETE FROM v1.project_fact_types WHERE id=%(id)s'

    PATCH_SQL = re.sub(r'\s+', ' ', """\
        UPDATE v1.project_fact_types
           SET last_modified_at=CURRENT_TIMESTAMP,
               last_modified_by=%(username)s,
               project_type_ids=%(project_type_ids)s,
               name=%(name)s,
               fact_type=%(fact_type)s,
               data_type=%(data_type)s,
               description=%(description)s,
               ui_options=%(ui_options)s,
               weight=%(weight)s
         WHERE id=%(id)s""")



import base64
import dataclasses
import typing
import urllib.parse

import sprockets.mixins.http
import yarl

from imbi import errors, oauth2, user
from imbi.clients import gitlab
from imbi.endpoints import base


@dataclasses.dataclass
class GitlabToken:
    access_token: str
    refresh_token: str


class RedirectHandler(sprockets.mixins.http.HTTPClientMixin,
                      base.RequestHandler):
    integration: 'oauth2.OAuth2Integration'

    NAME = 'gitlab-redirect'

    async def prepare(self) -> None:
        await super().prepare()
        if not self._finished:
            self.integration = await oauth2.OAuth2Integration.by_name(
                self.application, 'gitlab')
            if not self.integration:
                raise errors.IntegrationNotFound('gitlab')

    async def get(self):
        auth_code = self.get_query_argument('code')
        state = base64.b64decode(
            self.get_query_argument('state'), b'-_').decode('utf-8')
        username, target = state.rstrip('?').split(':')
        token = await self.exchange_code_for_token(auth_code)
        try:
            user_id, user_name, email = await self.fetch_gitlab_user(token)
            imbi_user = user.User(self.application, username=username)
            await imbi_user.refresh()
            if imbi_user.email_address != email:
                raise errors.Forbidden(
                    'mismatched user email: expected %r received %r',
                    imbi_user.email_address, email,
                    title='Gitlab authorization failure',
                    detail='unexpected email address {} for user {}'.format(
                        email, imbi_user.username))
            await self.integration.add_user_token(
                imbi_user, str(user_id), token.access_token,
                token.refresh_token)

        # Revoke the gitlab token if we cannot use or save it.  This
        # is a catch-all case since GitLab does not remove tokens based
        # on a lifetime.
        except Exception:
            await self.revoke_gitlab_token(token)
            raise

        target = yarl.URL(self.request.full_url()).with_path(target or '/ui/')
        self.redirect(str(target))

    async def exchange_code_for_token(self, code) -> GitlabToken:
        body = urllib.parse.urlencode({
            'client_id': self.integration.client_id,
            'client_secret': self.integration.client_secret,
            'grant_type': 'authorization_code',
            'redirect_uri': str(self.integration.callback_url),
            'code': code,
        })
        response = await self.http_fetch(
            str(self.integration.token_endpoint), method='POST', body=body,
            content_type='application/x-www-form-urlencoded')
        if not response.ok:
            self.logger.error('failed to exchange auth code for token: %s %s',
                              response.body['error'],
                              response.body['error_description'])
            raise errors.InternalServerError(
                'failed exchange auth code for token: %s', response.code,
                title='GitLab authorization failure',
                instance={
                    'error': response.body['error'],
                    'error_description': response.body['error_description'],
                })
        self.logger.debug('response_body %r', response.body)
        return GitlabToken(access_token=response.body['access_token'],
                           refresh_token=response.body['refresh_token'])

    async def fetch_gitlab_user(self, token: GitlabToken) \
            -> typing.Tuple[int, str, str]:
        response = await self.http_fetch(
            str(self.integration.api_endpoint / 'user'),
            request_headers={'Accept': 'application/json',
                             'Authorization': f'Bearer {token.access_token}'})
        if response.ok:
            return (response.body['id'], response.body['username'],
                    response.body['email'])
        raise errors.InternalServerError(
            'failed to retrieve GitLab user from access token',
            title='GitLab user lookup failure')

    async def revoke_gitlab_token(self, token: GitlabToken):
        ...


class GitLabIntegratedHandler(base.AuthenticatedRequestHandler):

    NAME = 'gitlab-integrated'
    client: gitlab.GitLabClient

    async def prepare(self) -> None:
        await super().prepare()
        if not self._finished:
            integration = await oauth2.OAuth2Integration.by_name(
                self.application, 'gitlab')
            if not integration:
                raise errors.IntegrationNotFound('gitlab')

            imbi_user = await self.get_current_user()  # never None
            tokens = await integration.get_user_tokens(imbi_user)
            if not tokens:
                raise errors.Forbidden('no GitLab tokens for %r', imbi_user,
                                       title='GitLab Not Connected')
            self.client = gitlab.GitLabClient(tokens[0], self.application)


class UserNamespacesHandler(GitLabIntegratedHandler):

    NAME = 'gitlab-user-namespace'

    async def get(self):
        entries = await self.client.fetch_all_pages(
            'groups', min_access_level=30)
        namespaces = [{'name': entry['full_name'], 'id': entry['id']}
                      for entry in entries]
        namespaces.sort(key=lambda elm: elm['name'])
        self.send_response(namespaces)


class ProjectsHandler(GitLabIntegratedHandler):

    NAME = 'gitlab-projects'

    async def get(self):
        group_arg = self.get_query_argument('group_id')
        try:
            group_id = int(group_arg)
        except ValueError:
            raise errors.BadRequest('invalid group ID: %r', group_arg,
                                    title='Invalid Query Parameter')

        projects = await self.client.fetch_all_pages(
            'groups', group_id, 'projects', include_subgroups='false',
            simple='true', with_shared='false')
        self.send_response([
            {
                'description': project['description'],
                'name': project['name'],
                'id': project['id'],
                'web_url': project['web_url'],
            }
            for project in projects
        ])



"""
Serve the OpenAPI related Endpoints

"""
import datetime

from imbi.endpoints import base


class RequestHandler(base.RequestHandler):
    """Services up the swagger related things rendered via Tornado's template
    system.

    """
    NAME = 'openapi'
    TTL = 300

    def get(self, *args):
        template = 'docs.html'
        if args and args[0] == 'openapi.yaml':
            template = 'openapi.yaml'
            self.set_header('Content-Type', 'text/yaml')
        stat = (self.application.settings['template_path'] / template).stat()
        self._add_last_modified_header(
            datetime.datetime.utcfromtimestamp(int(stat.st_mtime)))
        self._add_response_caching_headers(self.TTL)
        self.render(
            template, **{
                'host': self.request.host,
                'scheme': self.request.protocol,
                'settings': self.settings})



import re
from urllib import parse

import yarl

from imbi.endpoints import base


class _RequestHandlerMixin:

    ID_KEY = 'id'
    FIELDS = ['id', 'recorded_at', 'recorded_by', 'completed_at', 'project_id',
              'environment', 'change_type', 'description', 'link', 'notes',
              'ticket_slug', 'version']

    GET_SQL = re.sub(r'\s+', ' ', """\
        SELECT id, recorded_at, recorded_by, completed_at,
               project_id, environment, change_type, description,
               link, notes, ticket_slug, version
          FROM v1.operations_log
         WHERE id = %(id)s""")


class CollectionRequestHandler(_RequestHandlerMixin,
                               base.CollectionRequestHandler):

    NAME = 'operations-logs'
    ITEM_NAME = 'operations-log'

    COLLECTION_SQL = re.sub(r'\s+', ' ', """\
        SELECT o.id, o.recorded_at, o.recorded_by, o.completed_at,
               o.project_id, o.environment, o.change_type, o.description,
               o.link, o.notes, o.ticket_slug, o.version
          FROM v1.operations_log AS o
          {{JOIN}} {{WHERE}}
      ORDER BY o.recorded_at {{ORDER}}, o.id {{ORDER}}
         LIMIT %(limit)s""")

    VALUE_FILTER_CHUNK = re.sub(r'\s+', ' ', """\
        ((o.recorded_at = %(recorded_at_anchor)s AND o.id {{OP}} %(id_anchor)s)
        OR o.recorded_at {{OP}} %(recorded_at_anchor)s)""")

    FILTER_CHUNKS = {
        'from': 'o.recorded_at >= %(from)s',
        'to': 'o.recorded_at < %(to)s',
        'project_id': 'o.project_id = %(project_id)s',
        'namespace_id': 'p.namespace_id = %(namespace_id)s',
    }

    POST_SQL = re.sub(r'\s+', ' ', """\
        INSERT INTO v1.operations_log
                    (recorded_by, recorded_at, completed_at,
                     project_id, environment, change_type, description,
                     link, notes, ticket_slug, version)
             VALUES (%(recorded_by)s, %(recorded_at)s, %(completed_at)s,
                     %(project_id)s, %(environment)s, %(change_type)s,
                     %(description)s, %(link)s, %(notes)s, %(ticket_slug)s,
                     %(version)s)
          RETURNING id""")

    async def get(self, *args, **kwargs):
        kwargs['limit'] = int(self.get_query_argument('limit', '100')) + 1
        order = self.get_query_argument('order', 'desc')
        kwargs['from'] = self.get_query_argument('from', None)
        kwargs['to'] = self.get_query_argument('to', None)
        kwargs['namespace_id'] = self.get_query_argument('namespace_id', None)
        kwargs['project_id'] = self.get_query_argument('project_id', None)
        kwargs['recorded_at_anchor'] = self.get_query_argument(
            'recorded_at_anchor',
            None)
        kwargs['id_anchor'] = self.get_query_argument('id_anchor', None)
        page_direction = self.get_query_argument('page_direction', None)
        is_link = (page_direction is not None
                   and kwargs['recorded_at_anchor'] is not None
                   and kwargs['id_anchor'] is not None)

        join_sql = ''
        if kwargs['namespace_id'] is not None:
            join_sql += 'JOIN v1.projects AS p ON o.project_id = p.id'

        where_chunks = []
        for kwarg in self.FILTER_CHUNKS.keys():
            if kwargs[kwarg] is not None:
                where_chunks.append(self.FILTER_CHUNKS[kwarg])
        if is_link:
            if (order == 'asc' and page_direction == 'next') or \
                    (order == 'desc' and page_direction == 'previous'):
                op = '>'
            else:
                op = '<'
            where_chunks.append(self.VALUE_FILTER_CHUNK.replace('{{OP}}', op))
        where_sql = ''
        if where_chunks:
            where_sql = 'WHERE {}'.format(' AND '.join(where_chunks))

        if page_direction == 'previous':
            order = 'desc' if order == 'asc' else 'asc'

        sql = self.COLLECTION_SQL \
            .replace('{{JOIN}}', join_sql) \
            .replace('{{WHERE}}', where_sql) \
            .replace('{{ORDER}}', order)

        result = await self.postgres_execute(
            sql, kwargs, metric_name='get-{}'.format(self.NAME))
        rows = result.rows

        if page_direction == 'previous':
            rows = list(reversed(rows))

        next_needed, previous_needed = False, False
        if len(rows) == kwargs['limit']:
            if not is_link or page_direction == 'next':
                rows.pop(-1)
                next_needed = True
            else:
                rows.pop(0)
                previous_needed = True
        if page_direction == 'next':
            previous_needed = True
        elif page_direction == 'previous':
            next_needed = True

        request_url = yarl.URL(self.request.full_url())
        if next_needed:
            next_anchor = rows[-1]
            query = dict(request_url.query)
            query.update({
                'recorded_at_anchor': next_anchor['recorded_at'],
                'id_anchor': next_anchor['id'],
                'page_direction': 'next',
            })
            query_string = parse.urlencode(query)
            self._links['next'] = self.request.path + '?' + query_string
        if previous_needed:
            previous_anchor = rows[0]
            query = dict(request_url.query)
            query.update({
                'recorded_at_anchor': previous_anchor['recorded_at'],
                'id_anchor': previous_anchor['id'],
                'page_direction': 'previous',
            })
            query_string = parse.urlencode(query)
            self._links['previous'] = self.request.path + '?' + query_string

        self.send_response(rows)


class RecordRequestHandler(_RequestHandlerMixin,
                           base.AdminCRUDRequestHandler):
    NAME = 'operations-log'

    DELETE_SQL = 'DELETE FROM v1.operations_log WHERE id = %(id)s;'

    PATCH_SQL = re.sub(r'\s+', ' ', """\
        UPDATE v1.operations_log
           SET recorded_by = %(recorded_by)s,
               recorded_at = %(recorded_at)s,
               completed_at = %(completed_at)s,
               project_id = %(project_id)s,
               environment = %(environment)s,
               change_type = %(change_type)s,
               description = %(description)s,
               link = %(link)s,
               notes = %(notes)s,
               ticket_slug = %(ticket_slug)s,
               version = %(version)s
         WHERE id = %(id)s""")



import asyncio
import re

from imbi import errors, models
from imbi.endpoints import base
from imbi.opensearch import project


class _RequestHandlerMixin:

    ITEM_NAME = 'project'
    ID_KEY = ['id']
    FIELDS = ['id', 'namespace_id', 'project_type_id', 'name', 'slug',
              'description', 'environments', 'archived', 'gitlab_project_id',
              'sentry_project_slug', 'sonarqube_project_key',
              'pagerduty_service_id']
    TTL = 300

    GET_SQL = re.sub(r'\s+', ' ', """\
        SELECT a.id,
               a.created_at,
               a.created_by,
               a.last_modified_at,
               a.last_modified_by,
               a.namespace_id,
               b.name AS namespace,
               a.project_type_id,
               c.name AS project_type,
               a.name,
               a.slug,
               a.description,
               a.environments,
               a.archived,
               a.gitlab_project_id,
               a.sentry_project_slug,
               a.sonarqube_project_key,
               a.pagerduty_service_id
          FROM v1.projects AS a
          JOIN v1.namespaces AS b ON b.id = a.namespace_id
          JOIN v1.project_types AS c ON c.id = a.project_type_id
         WHERE a.id=%(id)s""")


class ProjectAttributeCollectionMixin(project.RequestHandlerMixin):

    async def post(self, *_args, **kwargs):
        result = await self._post(kwargs)
        await self.index_document(result['project_id'])


class ProjectAttributeCRUDMixin(project.RequestHandlerMixin):

    async def delete(self, *args, **kwargs):
        await super().delete(*args, **kwargs)
        await self.index_document(kwargs['project_id'])

    async def patch(self, *args, **kwargs):
        await super().patch(*args, **kwargs)
        await self.index_document(kwargs['project_id'])


class CollectionRequestHandler(project.RequestHandlerMixin,
                               _RequestHandlerMixin,
                               base.CollectionRequestHandler):
    NAME = 'projects'
    IS_COLLECTION = True
    COLLECTION_SQL = re.sub(r'\s+', ' ', """\
        SELECT a.id,
               a.created_at,
               a.created_by,
               a.last_modified_at,
               a.last_modified_by,
               a.namespace_id,
               b.name AS namespace,
               b.slug AS namespace_slug,
               b.icon_class AS namespace_icon,
               a.project_type_id,
               c.name AS project_type,
               c.icon_class AS project_icon,
               a.name,
               a.slug,
               a.description,
               a.environments,
               a.archived,
               a.gitlab_project_id,
               a.sentry_project_slug,
               a.sonarqube_project_key,
               a.pagerduty_service_id,
               v1.project_score(a.id) AS project_score
          FROM v1.projects AS a
          JOIN v1.namespaces AS b ON b.id = a.namespace_id
          JOIN v1.project_types AS c ON c.id = a.project_type_id
          {{WHERE}} {{ORDER_BY}} LIMIT %(limit)s OFFSET %(offset)s""")

    COUNT_SQL = re.sub(r'\s+', ' ', """\
        SELECT count(a.*) AS records
          FROM v1.projects AS a
          JOIN v1.namespaces AS b ON b.id = a.namespace_id
          JOIN v1.project_types AS c ON c.id = a.project_type_id
          {{WHERE}}""")

    FILTER_CHUNKS = {
        'name': 'to_tsvector(lower(a.name)) @@ websearch_to_tsquery(%(name)s)',
        'namespace_id': 'b.id = %(namespace_id)s',
        'project_type_id': 'c.id = %(project_type_id)s',
        'sonarqube_project_key':
            'a.sonarqube_project_key = %(sonarqube_project_key)s',
    }

    SORT_MAP = {
        'project_score': 'project_score',
        'namespace': 'b.name',
        'project_type': 'c.name',
        'name': 'a.name'
    }

    SORT_PATTERN = re.compile(
        r'(?:(?P<column>name|namespace|project_score|project_type) '
        r'(?P<direction>asc|desc))')

    POST_SQL = re.sub(r'\s+', ' ', """\
        INSERT INTO v1.projects
                    (namespace_id, project_type_id, created_by,  "name", slug,
                     description, environments)
             VALUES (%(namespace_id)s, %(project_type_id)s, %(username)s,
                     %(name)s, %(slug)s, %(description)s, %(environments)s)
          RETURNING id""")

    async def get(self, *args, **kwargs):
        kwargs['limit'] = int(self.get_query_argument('limit', '10'))
        kwargs['offset'] = int(self.get_query_argument('offset', '0'))
        where_chunks = []
        if self.get_query_argument('include_archived', 'false') == 'false':
            where_chunks.append('a.archived IS FALSE')
        for kwarg in self.FILTER_CHUNKS.keys():
            value = self.get_query_argument(kwarg, None)
            if value is not None:
                kwargs[kwarg] = value
                where_chunks.append(self.FILTER_CHUNKS[kwarg])
        where_sql = ''
        if where_chunks:
            where_sql = ' WHERE {}'.format(' AND '.join(where_chunks))
        sql = self.COLLECTION_SQL.replace('{{WHERE}}', where_sql)
        count_sql = self.COUNT_SQL.replace('{{WHERE}}', where_sql)

        order_sql = 'ORDER BY a.name ASC'
        order_by_chunks = []
        for match in self.SORT_PATTERN.finditer(
                self.get_query_argument('sort', '')):
            order_by_chunks.append(
                f'{match.group("column")} {match.group("direction").upper()}')
        if order_by_chunks:
            order_sql = ' ORDER BY {}'.format(', '.join(order_by_chunks))
        sql = sql.replace('{{ORDER_BY}}', order_sql)

        count = await self.postgres_execute(
            count_sql, kwargs, metric_name='count-{}'.format(self.NAME))
        result = await self.postgres_execute(
            sql, kwargs, metric_name='get-{}'.format(self.NAME))
        self.send_response({
            'rows': count.row['records'],
            'data': result.rows})

    async def post(self, *_args, **kwargs):
        result = await self._post(kwargs)
        await self.index_document(result['id'])


class RecordRequestHandler(project.RequestHandlerMixin,
                           _RequestHandlerMixin,
                           base.CRUDRequestHandler):

    NAME = 'project'

    DELETE_SQL = 'DELETE FROM v1.projects WHERE id=%(id)s'

    GET_FULL_SQL = re.sub(r'\s+', ' ', """\
        SELECT a.id,
               a.created_at,
               a.created_by,
               a.last_modified_at,
               a.last_modified_by,
               a.namespace_id,
               b.name AS namespace,
               b.slug AS namespace_slug,
               b.icon_class AS namespace_icon,
               a.project_type_id,
               c.name AS project_type,
               c.slug AS project_type_slug,
               c.icon_class AS project_icon,
               a.name,
               a.slug,
               a.description,
               a.environments,
               a.archived,
               a.gitlab_project_id,
               a.sentry_project_slug,
               a.sonarqube_project_key,
               a.pagerduty_service_id,
               v1.project_score(a.id)
          FROM v1.projects AS a
          JOIN v1.namespaces AS b ON b.id = a.namespace_id
          JOIN v1.project_types AS c ON c.id = a.project_type_id
         WHERE a.id=%(id)s""")

    GET_FACTS_SQL = re.sub(r'\s+', ' ', """\
        WITH project_type_id AS (SELECT project_type_id AS id
                                   FROM v1.projects
                                  WHERE id = %(id)s)
        SELECT a.id AS fact_type_id,
               a.name,
               b.recorded_at,
               b.recorded_by,
               CASE WHEN a.data_type = 'boolean' THEN b.value::bool::text
                    WHEN a.data_type = 'date' THEN b.value::date::text
                    WHEN a.data_type = 'decimal'
                         THEN b.value::numeric(9,2)::text
                    WHEN a.data_type = 'integer'
                         THEN b.value::integer::text
                    WHEN a.data_type = 'timestamp'
                         THEN b.value::timestamptz::text
                    ELSE b.value
                END AS value,
               a.data_type,
               a.fact_type,
               a.ui_options,
               CASE WHEN b.value IS NULL THEN 0
                    ELSE CASE WHEN a.fact_type = 'enum' THEN (
                                          SELECT score::NUMERIC(9,2)
                                            FROM v1.project_fact_type_enums
                                           WHERE fact_type_id = b.fact_type_id
                                             AND value = b.value)
                              WHEN a.fact_type = 'range' THEN (
                                          SELECT score::NUMERIC(9,2)
                                            FROM v1.project_fact_type_ranges
                                           WHERE fact_type_id = b.fact_type_id
                                             AND b.value::NUMERIC(9,2)
                                         BETWEEN min_value AND max_value)
                              WHEN a.data_type = 'boolean'
                               AND b.value = 'true' THEN 100
                              ELSE 0
                          END
                END AS score,
               CASE WHEN a.fact_type = 'enum' THEN (
                              SELECT icon_class
                                FROM v1.project_fact_type_enums
                               WHERE fact_type_id = b.fact_type_id
                                 AND value = b.value)
                    ELSE NULL
                END AS icon_class
          FROM v1.project_fact_types AS a
     LEFT JOIN v1.project_facts AS b
            ON b.fact_type_id = a.id
           AND b.project_id = %(id)s
         WHERE (SELECT id FROM project_type_id) = ANY(a.project_type_ids)
        ORDER BY a.name""")

    GET_LINKS_SQL = re.sub(r'\s+', ' ', """\
        SELECT a.link_type_id,
               b.link_type AS title,
               b.icon_class AS icon,
               a.url AS url
          FROM v1.project_links AS a
          JOIN v1.project_link_types AS b ON b.id = a.link_type_id
         WHERE a.project_id=%(id)s
         ORDER BY b.link_type""")

    GET_URLS_SQL = re.sub(r'\s+', ' ', """\
        SELECT environment, url
          FROM v1.project_urls
         WHERE project_id=%(id)s
         ORDER BY environment""")

    PATCH_SQL = re.sub(r'\s+', ' ', """\
        UPDATE v1.projects
           SET namespace_id=%(namespace_id)s,
               project_type_id=%(project_type_id)s,
               last_modified_at=CURRENT_TIMESTAMP,
               last_modified_by=%(username)s,
               "name"=%(name)s,
               slug=%(slug)s,
               description=%(description)s,
               environments=%(environments)s,
               archived=%(archived)s,
               gitlab_project_id=%(gitlab_project_id)s,
               sentry_project_slug=%(sentry_project_slug)s,
               sonarqube_project_key=%(sonarqube_project_key)s,
               pagerduty_service_id=%(pagerduty_service_id)s
         WHERE id=%(id)s""")

    async def delete(self, *args, **kwargs):
        await super().delete(*args, **kwargs)
        await self.search_index.delete_document(kwargs['id'])

    async def get(self, *args, **kwargs):
        if self.get_argument('full', 'false') == 'true':
            query_args = self._get_query_kwargs(kwargs)
            project, facts, links, urls = await asyncio.gather(
                self.postgres_execute(
                    self.GET_FULL_SQL, query_args, 'get-{}'.format(self.NAME)),
                self.postgres_execute(
                    self.GET_FACTS_SQL, query_args, 'get-project-facts'),
                self.postgres_execute(
                    self.GET_LINKS_SQL, query_args, 'get-project-links'),
                self.postgres_execute(
                    self.GET_URLS_SQL, query_args, 'get-project-urls'))

            if not project.row_count or not project.row:
                raise errors.ItemNotFound()

            output = project.row
            output.update({
                'facts': facts.rows,
                'links': links.rows,
                'urls': {row['environment']: row['url'] for row in urls.rows}
            })
            self.send_response(output)
        else:
            await self._get(kwargs)

    async def patch(self, *args, **kwargs):
        await super().patch(*args, **kwargs)
        await self.index_document(kwargs['id'])


class SearchRequestHandler(project.RequestHandlerMixin,
                           base.AuthenticatedRequestHandler):

    async def get(self):
        result = await self.search_index.search(self.get_query_argument('s'))
        self.send_response(result)


class SearchIndexRequestHandler(project.RequestHandlerMixin,
                                base.AuthenticatedRequestHandler):

    SQL = re.sub(r'\s+', ' ', """\
        SELECT id
          FROM v1.projects
         ORDER BY id""")

    async def get(self):
        result = await self.postgres_execute(self.SQL)
        ids = [row['id'] for row in result]
        for project_id in ids:
            value = await models.project(project_id, self.application)
            await self.search_index.index_document(value)

        self.send_response({
            'status': 'ok',
            'message': f'Queued {len(ids)} projects for indexing'})



import re
import typing

import sprockets_postgres
import yarl

from imbi import models
from imbi.automations import base
from imbi.clients import gitlab, sonarqube
if typing.TYPE_CHECKING:
    from imbi import app, user


class SonarCreateProject(base.Automation):

    INSERT_SQL = re.sub(
        r'\s+', ' ', """\
        INSERT INTO v1.project_links
                    (project_id, link_type_id, created_by, url)
             VALUES (%(project_id)s, %(link_type_id)s,
                     %(username)s, %(url)s)""")
    UPDATE_SQL = re.sub(
        r'\s+', ' ', """
        UPDATE v1.projects
           SET sonarqube_project_key = %(sonar_key)s
         WHERE id = %(project_id)s""")

    def __init__(self,
                 application: 'app.Application',
                 project_id: int,
                 public_url: yarl.URL,
                 current_user: 'user.User',
                 db: sprockets_postgres.PostgresConnector):
        super().__init__(application, current_user, db)
        self._public_url = public_url
        self._imbi_project_id = project_id
        self._gitlab_project: typing.Optional[gitlab.ProjectInfo] = None
        self._project: typing.Optional[models.Project] = None
        self._sonar_client: typing.Optional[sonarqube.SonarQubeClient] = None

    async def prepare(self) -> typing.List[str]:
        self._project = await self._get_project(self._imbi_project_id)
        token = await self._get_gitlab_token()
        if self._project and self._project.gitlab_project_id is None:
            self._add_error('GitLab project does not exist for {}',
                            self._project.slug)
        if not self._has_error():
            client = gitlab.GitLabClient(token, self.application)
            self._gitlab_project = await client.fetch_project(
                self._project.gitlab_project_id)
            if self._gitlab_project is None:
                self._add_error('GitLab project id {} does not exist',
                                self._project.gitlab_project_id)

        if not self._has_error():
            self._sonar_client = sonarqube.SonarQubeClient(self.application)
            if not self._sonar_client.enabled:
                self._add_error('SonarQube integration is not configured')

        return self.errors

    async def run(self) -> dict:
        project_key, dashboard = await self._sonar_client.create_project(
            self._project,
            public_url=self._public_url,
            main_branch_name=self._gitlab_project.default_branch)

        link_id = self.automation_settings['sonarqube'].get(
            'project_link_type_id')
        if link_id:
            await self.db.execute(
                self.INSERT_SQL, {
                    'link_type_id': link_id,
                    'project_id': self._project.id,
                    'url': dashboard,
                    'username': self.user.username
                })
            await self.db.execute(self.UPDATE_SQL, {
                'project_id': self._project.id,
                'sonar_key': project_key
            })
        return {
            'sonarqube_dashboard': dashboard,
            'sonarqube_project_key': project_key
        }



from imbi.endpoints import base


class LoginRequestHandler(base.RequestHandler):

    NAME = 'ui-login'

    async def post(self, *args, **kwargs):
        body = self.get_request_body()
        if not await self.session.authenticate(
                body.get('username'), body.get('password')):
            self.logger.debug('Session failed to authenticate')
            self.set_status(401)
            self.send_response({'message': 'Authentication Failure'})
            return
        await self.session.save()
        self.set_status(200)
        self.send_response(self.session.user.as_dict())


class LogoutRequestHandler(base.RequestHandler):

    NAME = 'ui-logout'

    async def get(self, *args, **kwargs):
        await self.session.clear()
        self.send_response({'loggedOut': True})



import datetime
import json
import logging

import iso8601
import openapi_spec_validator
import tornado_openapi3
import umsgpack
import validators
import yaml
from jsonschema import validators as jsonschema_validators
from openapi_core.schema.specs import factories, models
from tornado import template

from imbi import transcoders

LOGGER = logging.getLogger(__name__)


class EMailFormatter:

    @staticmethod
    def validate(value) -> bool:
        return validators.email(value)

    @staticmethod
    def unmarshal(value: str) -> str:
        return value


class ISO8601Formatter:

    @staticmethod
    def validate(value) -> bool:
        try:
            iso8601.parse_date(value)
        except iso8601.ParseError:
            return False
        return True

    @staticmethod
    def unmarshal(value: str) -> datetime.datetime:
        return iso8601.parse_date(value)


class URIFormatter:

    @staticmethod
    def validate(value) -> bool:
        return validators.url(value)

    @staticmethod
    def unmarshal(value: str) -> str:
        return value


_openapi_formatters = {
    'email': EMailFormatter,
    'iso8601-timestamp': ISO8601Formatter,
    'uri': URIFormatter,
}
_openapi_deserializers = {
    'application/json-patch+json': json.loads,
    'application/json-patch+msgpack': umsgpack.unpackb,
    'application/msgpack': umsgpack.unpackb,
    'application/problem+json': json.loads,
    'application/problem+msgpack': umsgpack.unpackb,
    'application/x-www-form-urlencoded': transcoders.parse_form_body,
    'application/yaml': yaml.safe_load,
}

# This is updated in `create_spec` the first time that it is called.
# "uncaching" this causes a 10x performance hit in testing
_openapi_spec_dict = {}


def create_spec(settings: dict) -> models.Spec:
    """Create and return the OpenAPI v3 Spec Model"""
    if not _openapi_spec_dict:
        loader = template.Loader(str(settings['template_path']))
        _openapi_spec_dict.update(yaml.safe_load(
            loader.load('openapi.yaml').generate(settings=settings)))
        # Remove servers for validation to prevent hostname validation errors
        _openapi_spec_dict.pop('servers', None)

    spec_resolver = jsonschema_validators.RefResolver(
        '', _openapi_spec_dict,
        handlers=openapi_spec_validator.default_handlers)
    spec_factory = factories.SpecFactory(
        spec_resolver, config={'validate_spec': False})
    return spec_factory.create(_openapi_spec_dict, spec_url='')


def request_validator(settings: dict) -> tornado_openapi3.RequestValidator:
    return tornado_openapi3.RequestValidator(
        spec=create_spec(settings),
        custom_formatters=_openapi_formatters,
        custom_media_type_deserializers=_openapi_deserializers)



from tornado import web

from imbi import constants
from . import (activity_feed, authentication_tokens, cookie_cutters, dashboard,
               environments, fact_type_enums, fact_type_ranges, fact_types,
               groups, integrations, metrics, namespaces, openapi, opensearch,
               operations_log, permissions, project_activity_feed,
               project_dependencies, project_fact_history, project_fact_types,
               project_facts, project_link_types, project_links,
               project_notes, project_score_history, project_secrets,
               project_types, project_urls, projects, reports, status, ui)

URLS = [
    web.url(r'^/$', ui.IndexRequestHandler),
    web.url(r'^/activity-feed$', activity_feed.RequestHandler),
    web.url(r'^/api-docs/$', openapi.RequestHandler),
    web.url(r'^/api-docs/(openapi.yaml)$', openapi.RequestHandler),
    web.url(r'^/authentication-tokens$', authentication_tokens.RequestHandler),
    web.url(r'^/authentication-tokens/(?P<token>{})$'.format(
                constants.UUID_PATTERN),
            authentication_tokens.RequestHandler),
    web.url(r'^/cookie-cutters$', cookie_cutters.CollectionRequestHandler),
    web.url(r'^/cookie-cutters/(?P<name>[\w_\-%\+]+)$',
            cookie_cutters.RecordRequestHandler,
            name='cookie-cutter'),
    web.url(r'^/dashboard$', dashboard.RequestHandler),
    web.url(r'^/environments$', environments.CollectionRequestHandler),
    web.url(r'^/environments/(?P<name>[\w_\-%\+]+)$',
            environments.RecordRequestHandler,
            name='environment'),
    web.url(r'^/groups$', groups.CollectionRequestHandler),
    web.url(r'^/groups/(?P<name>[\w_\-%\+]+)$',
            groups.RecordRequestHandler,
            name='group'),
    web.url(r'^/metrics$', metrics.RequestHandler),
    web.url(r'^/namespaces$', namespaces.CollectionRequestHandler),
    web.url(r'^/namespaces/(?P<id>\d+)$',
            namespaces.RecordRequestHandler,
            name='namespace'),
    web.url(r'^/opensearch/(?P<index>[\w_\-%\+]+)$',
            opensearch.RequestHandler),
    web.url(r'^/operations-log$', operations_log.CollectionRequestHandler),
    web.url(r'^/operations-log/(?P<id>\d+)$',
            operations_log.RecordRequestHandler,
            name='operations-log'),
    web.url(r'^/permissions$', permissions.RequestHandler),
    web.url(r'^/project-fact-types$', fact_types.CollectionRequestHandler),
    web.url(r'^/project-fact-types/(?P<id>\d+)$',
            fact_types.RecordRequestHandler, name='fact-type'),
    web.url(r'^/project-fact-type-enums$',
            fact_type_enums.CollectionRequestHandler),
    web.url(r'^/project-fact-type-enums/(?P<id>\d+)$',
            fact_type_enums.RecordRequestHandler,
            name='fact-type-enum'),
    web.url(r'^/project-fact-type-ranges$',
            fact_type_ranges.CollectionRequestHandler),
    web.url(r'^/project-fact-type-ranges/(?P<id>\d+)$',
            fact_type_ranges.RecordRequestHandler,
            name='fact-type-range'),
    web.url(r'^/project-link-types$',
            project_link_types.CollectionRequestHandler),
    web.url(r'^/project-link-types/(?P<id>\d+)$',
            project_link_types.RecordRequestHandler,
            name='project-link-type'),
    web.url(r'^/project-types$', project_types.CollectionRequestHandler),
    web.url(r'^/project-types/(?P<id>\d+)$',
            project_types.RecordRequestHandler,
            name='project-type'),
    web.url(r'^/projects/build-search-index',
            projects.SearchIndexRequestHandler),
    web.url(r'^/projects$', projects.CollectionRequestHandler,
            name='projects'),
    web.url(r'^/projects/(?P<id>\d+)$',
            projects.RecordRequestHandler,
            name='project'),
    web.url(r'^/projects/(?P<project_id>\d+)/dependencies$',
            project_dependencies.CollectionRequestHandler,
            name='project-dependencies'),
    web.url(r'^/projects/(?P<project_id>\d+)/dependencies/'
            r'(?P<dependency_id>\d+)$',
            project_dependencies.RecordRequestHandler,
            name='project-dependency'),
    web.url(r'^/projects/(?P<project_id>\d+)/facts$',
            project_facts.CollectionRequestHandler,
            name='project-facts'),
    web.url(r'^/projects/(?P<project_id>\d+)/fact-types$',
            project_fact_types.CollectionRequestHandler,
            name='project-fact-types'),
    web.url(r'^/projects/(?P<project_id>\d+)/fact-history$',
            project_fact_history.CollectionRequestHandler,
            name='project-fact-history'),
    web.url(r'^/projects/(?P<project_id>\d+)/feed$',
            project_activity_feed.CollectionRequestHandler,
            name='project-activity-feed'),
    web.url(r'^/projects/(?P<project_id>\d+)/links$',
            project_links.CollectionRequestHandler,
            name='project-links'),
    web.url(r'^/projects/(?P<project_id>\d+)/links/'
            r'(?P<link_type_id>\d+)$',
            project_links.RecordRequestHandler,
            name='project-link'),
    web.url(r'^/projects/(?P<project_id>\d+)/notes$',
            project_notes.CollectionRequestHandler,
            name='project-notes'),
    web.url(r'^/projects/(?P<project_id>\d+)/notes/(?P<id>\d+)$',
            project_notes.RecordHandler,
            name='project-note'),
    web.url(r'^/projects/(?P<project_id>\d+)/score-history$',
            project_score_history.CollectionRequestHandler,
            name='project-score-history'),
    web.url(r'^/projects/(?P<project_id>\d+)/secrets$',
            project_secrets.CollectionRequestHandler,
            name='project-secrets'),
    web.url(r'^/projects/(?P<project_id>\d+)/urls$',
            project_urls.CollectionRequestHandler,
            name='project-urls'),
    web.url(r'^/projects/(?P<project_id>\d+)/urls/'
            r'(?P<environment>[\w_\-%\+]+)$',
            project_urls.RecordRequestHandler,
            name='project-url'),
    web.url(r'^/status$', status.RequestHandler),
] + integrations.URLS + ui.URLS + reports.URLS



"""
System Reports

"""
from tornado import web

from . import gitlab, sentry, sonarqube

URLS = [
    web.url(r'^/ui/automations/gitlab/commit',
            gitlab.InitialCommitRequestHandler),
    web.url(r'^/ui/automations/gitlab/create', gitlab.CreationRequestHandler),
    web.url(r'^/ui/automations/sentry/create',
            sentry.ProjectCreationRequestHandler),
    web.url(r'^/ui/automations/sonarqube/create',
            sonarqube.CreationRequestHandler),
]



import asyncio

from imbi.endpoints import base, cookie_cutters, environments, groups, \
    namespaces, fact_types, project_link_types, project_types
from imbi.endpoints.integrations import oauth2
from imbi.opensearch import project


class RequestHandler(base.RequestHandler):

    NAME = 'metadata'

    async def get(self) -> None:
        """Return all metadata in a single request"""
        project_index = project.ProjectIndex(self.application)
        results = await asyncio.gather(
            self.postgres_execute(
                cookie_cutters.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='cookie-cutters'),
            self.postgres_execute(
                environments.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='environments'),
            self.postgres_execute(
                groups.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='groups'),
            self.postgres_execute(
                namespaces.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='namespaces'),
            self.postgres_execute(
                fact_types.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='project-fact-types'),
            self.postgres_execute(
                project_link_types.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='project-link-types'),
            self.postgres_execute(
                project_types.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='project-types'),
            project_index.searchable_fields(),
            self.postgres_execute(
                oauth2.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='integrations'),
        )

        automations = self.settings['automations']

        oauth_details = results[8].rows
        gitlab = {'enabled': False}
        cfg = automations.get('gitlab')
        if cfg:
            gitlab['project_link_type_id'] = cfg.get('project_link_type_id')
            gitlab_auth = [r for r in oauth_details if r['name'] == 'gitlab']
            if gitlab_auth:
                gitlab_auth = gitlab_auth[0]
                gitlab.update({
                    'authorizationEndpoint':
                        gitlab_auth['authorization_endpoint'],
                    'clientId': gitlab_auth['client_id'],
                    'redirectURI': gitlab_auth['callback_url'],
                })
            gitlab['enabled'] = (cfg['enabled'] and
                                 gitlab.get('clientId') is not None)

        sentry = {'enabled': False}
        cfg = automations.get('sentry')
        if cfg:
            sentry['project_link_type_id'] = cfg.get('project_link_type_id')
            sentry['enabled'] = cfg['enabled']

        self.send_response({
            'integrations': {
                'grafana': {
                    'enabled': automations['grafana']['enabled'],
                    'project_link_type_id':
                        automations['grafana']['project_link_type_id']
                },
                'gitlab': gitlab,
                'sentry': {
                    'enabled': automations['sentry']['enabled'],
                    'project_link_type_id':
                        automations['sentry']['project_link_type_id']
                },
                'sonarqube': {
                    'enabled': automations['sonarqube']['enabled'],
                    'project_link_type_id':
                        automations['sonarqube']['project_link_type_id']
                }
            },
            'metadata': {
                'cookie_cutters': results[0].rows,
                'environments': results[1].rows,
                'groups': results[2].rows,
                'namespaces': results[3].rows,
                'project_fact_types': results[4].rows,
                'project_link_types': results[5].rows,
                'project_types': results[6].rows
            },
            'opensearch': {
                'fields': results[7]
            },
            'project_url_template': self.settings['project_url_template']
        })



import re

from imbi.endpoints import base


class CollectionRequestHandler(base.CollectionRequestHandler):

    NAME = 'project-fact-history'
    ID = 'project_id'
    COLLECTION_SQL = re.sub(r'\s+', ' ', """\
        SELECT a.fact_type_id,
               a.recorded_at,
               a.value,
               a.score,
               a.weight,
               b.icon_class,
               c.display_name AS recorded_by
          FROM v1.project_fact_history AS a
          LEFT JOIN v1.project_fact_type_enums AS b
            ON a.fact_type_id = b.fact_type_id
           AND a.value = b.value
          LEFT JOIN v1.users as c
            ON a.recorded_by = c.username
         WHERE a.project_id = %(project_id)s
         ORDER BY a.recorded_at DESC
         LIMIT %(limit)s OFFSET %(offset)s""")

    async def get(self, *args, **kwargs):
        kwargs['limit'] = int(self.get_query_argument('limit', '25'))
        kwargs['offset'] = int(self.get_query_argument('offset', '0'))

        result = await self.postgres_execute(
            self.COLLECTION_SQL, kwargs,
            metric_name='get-{}'.format(self.NAME))
        self.send_response(result.rows)



"""
Core Application
================

"""
import asyncio
import datetime
import json
import logging
import os
import re
import typing

import aioredis
import sprockets_postgres
try:
    import sentry_sdk
except ImportError:
    sentry_sdk = None
import umsgpack
from sprockets import http
from sprockets.http import app
from sprockets.mixins.mediatype import content
from tornado import httputil, ioloop, web
try:
    from sentry_sdk.integrations import logging as sentry_logging
    from sentry_sdk.integrations import tornado as sentry_tornado
except ImportError:
    sentry_logging, sentry_tornado = None, None

from imbi import (cors, endpoints, errors, keychain, openapi,
                  permissions, stats, transcoders, version)
from imbi.clients import opensearch
from imbi.endpoints import default

LOGGER = logging.getLogger(__name__)

DEFAULT_SESSION_POOL_SIZE = 10
REQUEST_LOG_FORMAT = '%d %s %.2fms %s'
SIGNED_VALUE_PATTERN = re.compile(r'^(?:[1-9][0-9]*)\|(?:.*)$')


class Application(sprockets_postgres.ApplicationMixin, app.Application):

    def __init__(self, **settings):
        LOGGER.info('imbi v%s starting', settings['version'])
        settings['default_handler_class'] = default.RequestHandler
        settings['permissions'] = permissions.PERMISSIONS
        super(Application, self).__init__(endpoints.URLS, **settings)
        self._ready_to_serve = False
        self._request_logger = logging.getLogger('imbi')
        self.keychain = keychain.Keychain(self.settings['encryption_key'])
        self.loop: typing.Optional[ioloop.IOLoop] = None
        self.on_start_callbacks.append(self.on_start)
        self.on_shutdown_callbacks.append(self.on_shutdown)
        self.openapi_validator = openapi.request_validator(self.settings)
        self.opensearch: typing.Optional[opensearch.OpenSearch] = None
        self.session_redis: typing.Optional[aioredis.Redis] = None
        self.started_at = datetime.datetime.now(datetime.timezone.utc)
        self.started_at_str = self.started_at.isoformat()
        self.startup_complete: typing.Optional[asyncio.Event] = None
        self.stats: typing.Optional[stats.Stats] = None

        content.set_default_content_type(self, 'application/json')
        content.add_transcoder(self, transcoders.FormTranscoder())
        content.add_transcoder(self, transcoders.JSONTranscoder())
        content.add_transcoder(self, transcoders.MsgPackTranscoder())
        content.add_text_content_type(
            self, 'application/json-patch+json', 'utf-8',
            json.dumps, json.loads)
        content.add_binary_content_type(
            self, 'application/json-patch+msgpack',
            umsgpack.packb, umsgpack.unpackb)

        errors.set_canonical_server(self.settings['canonical_server_name'])

        if self.settings.get('cors'):
            kwargs = {
                name: self.settings['cors'][name]
                for name in {'allow_any_origin', 'allow_credentials',
                             'allow_methods', 'exposed_headers',
                             'max_age'}
                if name in self.settings['cors']
            }
            self.cors_config = cors.CORSConfig(**kwargs)
        else:
            self.cors_config = cors.CORSConfig()
        for allowed_origin in self.settings.get('cors', {}).get('origins', []):
            self.cors_config.allowed_origins.add(allowed_origin)

    def hash_password(self, password: str) -> str:
        """Generate a HMAC-SHA512 hash of `password`."""
        hashed_bytes = self.keychain.hash(password)
        return '{}:{}'.format(self.keychain.algorithm.name, hashed_bytes.hex())

    def decrypt_value(self, value: str) -> str:
        """Decrypt a value that is encrypted using `encrypt_value`."""
        plaintext_bytes = self.keychain.decrypt(bytes.fromhex(value))
        return plaintext_bytes.decode('utf-8')

    def encrypt_value(self, value: str) -> str:
        """Encrypt a value."""
        cipher_bytes = self.keychain.encrypt(value.encode('utf-8'))
        return cipher_bytes.hex()

    def log_request(self, handler: web.RequestHandler) -> None:
        """Writes a completed HTTP request to the logs"""
        if handler.request.path == '/status':
            return
        request_time = 1000.0 * handler.request.request_time()
        status_code = handler.get_status()
        if status_code < 400:
            self._request_logger.info(
                REQUEST_LOG_FORMAT, status_code, handler._request_summary(),
                request_time, handler.request.headers.get('User-Agent'))
        if 400 <= status_code < 500:
            self._request_logger.warning(
                REQUEST_LOG_FORMAT, status_code, handler._request_summary(),
                request_time, handler.request.headers.get('User-Agent'))
        if status_code >= 500:
            self._request_logger.error(
                REQUEST_LOG_FORMAT, status_code, handler._request_summary(),
                request_time, handler.request.headers.get('User-Agent'))

    async def on_start(self,
                       _app: http.app.Application,
                       _loop: ioloop.IOLoop) -> None:

        """Invoked on startup of the application"""
        self.startup_complete = asyncio.Event()
        if sentry_sdk and self.settings.get('sentry_backend_dsn'):
            sentry_sdk.init(
                debug=self.settings['debug'],
                dsn=self.settings['sentry_backend_dsn'],
                environment=os.environ.get('environment', 'production'),
                integrations=[
                    sentry_tornado.TornadoIntegration(),
                    sentry_logging.LoggingIntegration(
                        event_level=logging.CRITICAL)
                    ],
                release=version)

        self.loop = ioloop.IOLoop.current()
        try:
            self.session_redis = aioredis.Redis(
                await aioredis.create_pool(
                    self.settings['session_redis_url'],
                    maxsize=self.settings['session_pool_size']))
        except (OSError, ConnectionRefusedError) as error:
            LOGGER.info('Error connecting to Session redis: %r', error)
            self.stop(self.loop)
            return

        try:
            pool = aioredis.Redis(
                await aioredis.create_pool(
                    self.settings['stats_redis_url'],
                    maxsize=self.settings['stats_pool_size']))
        except (OSError, ConnectionRefusedError) as error:
            LOGGER.info('Error connecting to Stats redis: %r', error)
            self.stop(self.loop)
            return
        else:
            self.stats = stats.Stats(pool)

        await self._postgres_connected.wait()

        self.opensearch = opensearch.OpenSearch(self.settings['opensearch'])
        if not await self.opensearch.initialize():
            self.stop(self.loop)
            return

        self.startup_complete.set()
        self._ready_to_serve = True
        LOGGER.info('Application startup complete')

    async def on_shutdown(self, *_args, **_kwargs) -> None:
        await self.opensearch.stop()

    def validate_request(self, request: httputil.HTTPServerRequest) -> None:
        """Validate the inbound request, raising any number of OpenAPI
        exceptions on error.

        """
        self.openapi_validator.validate(request).raise_for_errors()

    @property
    def ready_to_serve(self) -> bool:
        """Indicates if the service is available to respond to HTTP requests"""
        return self._ready_to_serve



import re

from tornado import web

from imbi.endpoints import base


class RequestHandler(base.RequestHandler):

    NAME = 'reports-namespace-kpis'

    SQL = re.sub(r'\s+', ' ', """\
        WITH projects_with_facts AS (
            SELECT a.id, a.namespace_id
              FROM v1.projects AS a
         LEFT JOIN v1.project_fact_types AS b
                ON a.project_type_id = ANY(b.project_type_ids)
             WHERE a.archived IS NOT TRUE
               AND b.weight > 0
          GROUP BY a.id, a.namespace_id
            HAVING count(b.*) > 0),
        project_scores AS (
            SELECT a.namespace_id,
                   b.name AS namespace,
                   a.id,
                   v1.project_score(a.id)
              FROM projects_with_facts AS a
              JOIN v1.namespaces AS b
                ON b.id = a.namespace_id)
        SELECT namespace_id,
               namespace,
               count(*) AS projects,
               avg(project_score) AS stack_health_score,
               sum(project_score)::INT AS total_project_score,
               count(*) * 100 AS total_possible_project_score,
               (((sum(project_score)::NUMERIC(9,2)
                   / (count(*) * 100)::NUMERIC(9,2))
                 * 100)::NUMERIC(9,2))::TEXT || '%' AS percent_of_tpps
          FROM project_scores
         GROUP BY namespace_id, namespace
         ORDER BY namespace;""")

    @web.authenticated
    async def get(self):
        result = await self.postgres_execute(self.SQL, metric_name=self.NAME)
        self.send_response(result.rows)



import re

from imbi.endpoints import base


class CollectionRequestHandler(base.CollectionRequestHandler):

    NAME = 'project-score-history'
    ID = 'project_id'
    COLLECTION_SQL = re.sub(r'\s+', ' ', """\
        WITH history AS (
                SELECT changed_at::DATE AS "date",
                       max(score::INT) AS score
                  FROM v1.project_score_history
                 WHERE project_id = %(project_id)s
              GROUP BY changed_at::DATE),
             current AS (
                SELECT CURRENT_DATE AS "date",
                       v1.project_score(%(project_id)s)::INT AS score),
             records AS (
                SELECT "date", score
                  FROM history
                 UNION
                 SELECT CURRENT_DATE AS "date",
                        v1.project_score(%(project_id)s)::INT AS score)
            SELECT "date", max(score) AS score
              FROM records
          GROUP BY "date"
          ORDER BY "date" ASC;""")



import dataclasses
import logging
import re
import typing

import yarl

from imbi import errors
if typing.TYPE_CHECKING:
    from imbi import app, user

LOGGER = logging.getLogger(__name__)
UNSET_URL = yarl.URL()


@dataclasses.dataclass
class IntegrationToken:
    integration: 'OAuth2Integration'
    access_token: str
    refresh_token: str
    external_id: str


class OAuth2Integration:
    SQL_REFRESH = re.sub(r'\s+', ' ', """\
        SELECT api_endpoint, authorization_endpoint, token_endpoint,
               revoke_endpoint, client_id, client_secret, public_client,
               callback_url
          FROM v1.oauth2_integrations
         WHERE name = %(name)s""")

    SQL_ADD_TOKEN = re.sub(r'\s+', ' ', """\
        INSERT INTO v1.user_oauth2_tokens(username, integration, external_id,
                                          access_token, refresh_token)
             VALUES (%(username)s, %(integration)s, %(external_id)s,
                     %(access_token)s, %(refresh_token)s)""")

    SQL_GET_TOKENS = re.sub(r'\s+', ' ', """\
        SELECT access_token, refresh_token, external_id
          FROM v1.user_oauth2_tokens
         WHERE username = %(username)s
           AND integration = %(integration)s""")

    def __init__(self,
                 application: 'app.Application',
                 name: str) -> None:
        self._application = application
        self.logger = LOGGER.getChild(self.__class__.__name__)
        self.name = name
        self.authorization_endpoint: yarl.URL = UNSET_URL
        self.api_endpoint: typing.Optional[yarl.URL] = None
        self.token_endpoint: yarl.URL = UNSET_URL
        self.revoke_endpoint: typing.Optional[yarl.URL] = None
        self.client_id: str = ''
        self.client_secret: str = ''
        self.public_client: bool = True
        self.callback_url: typing.Optional[yarl.URL] = UNSET_URL

    @classmethod
    async def by_name(cls,
                      application: 'app.Application',
                      name: str) -> typing.Optional['OAuth2Integration']:
        instance = cls(application, name)
        await instance.refresh()
        return instance if instance.is_valid else None

    async def refresh(self) -> None:
        async with self._application.postgres_connector(
                on_error=self._on_postgres_error) as conn:
            result = await conn.execute(self.SQL_REFRESH, {'name': self.name})

        if result:
            self.authorization_endpoint = yarl.URL(
                result.row['authorization_endpoint'])
            self.token_endpoint = yarl.URL(result.row['token_endpoint'])
            self.client_id = result.row['client_id']
            self.client_secret = result.row['client_secret']
            self.public_client = result.row['public_client']

            if result.row['api_endpoint']:
                self.api_endpoint = yarl.URL(result.row['api_endpoint'])
            else:
                self.api_endpoint = None

            if result.row['revoke_endpoint']:
                self.revoke_endpoint = yarl.URL(result.row['revoke_endpoint'])
            else:
                self.revoke_endpoint = None

            if result.row['callback_url']:
                self.callback_url = yarl.URL(result.row['callback_url'])
            else:
                self.callback_url = None
        else:
            self._reset()

    async def add_user_token(self,
                             user_info: 'user.User',
                             external_id: str,
                             access_token: str,
                             refresh_token: str) -> None:
        async with self._application.postgres_connector(
                on_error=self._on_postgres_error) as conn:
            await conn.execute(
                self.SQL_ADD_TOKEN,
                {
                    'integration': self.name,
                    'username': user_info.username,
                    'external_id': external_id,
                    'access_token': access_token,
                    'refresh_token': refresh_token
                })

    async def get_user_tokens(self, user_info: 'user.User') \
            -> typing.List[IntegrationToken]:
        async with self._application.postgres_connector(
                on_error=self._on_postgres_error) as conn:
            result = await conn.execute(self.SQL_GET_TOKENS, {
                'integration': self.name, 'username': user_info.username})
        return [
            IntegrationToken(
                self,  row['access_token'], row['refresh_token'],
                row['external_id'])
            for row in result
        ]

    @property
    def is_valid(self) -> bool:
        return (self.authorization_endpoint != UNSET_URL and
                self.token_endpoint != UNSET_URL and
                self.client_id and (self.client_secret or self.public_client))

    @staticmethod
    def _on_postgres_error(_metric_name: str, exc: Exception) -> None:
        raise errors.DatabaseError(error=exc)

    def _reset(self) -> None:
        self.api_endpoint = None
        self.authorization_endpoint = UNSET_URL
        self.callback_url = UNSET_URL
        self.token_endpoint = UNSET_URL
        self.revoke_endpoint = None
        self.client_id = ''
        self.client_secret = ''
        self.public_client = True



import re

from imbi.endpoints import base


class _RequestHandlerMixin:

    ID_KEY = ['id']
    FIELDS = ['id', 'fact_type_id', 'value', 'icon_class',  'score']

    GET_SQL = re.sub(r'\s+', ' ', """\
        SELECT id, fact_type_id, created_at, created_by,
               last_modified_at, last_modified_by, value, icon_class, score
          FROM v1.project_fact_type_enums
         WHERE id=%(id)s""")


class CollectionRequestHandler(_RequestHandlerMixin,
                               base.CollectionRequestHandler):

    NAME = 'fact-type-enums'
    ITEM_NAME = 'fact-type-enum'

    COLLECTION_SQL = re.sub(r'\s+', ' ', """\
        SELECT a.id, a.fact_type_id, a.value, a.icon_class, a.score
          FROM v1.project_fact_type_enums AS a
          JOIN v1.project_fact_types AS b
           ON b.id = a.fact_type_id
      ORDER BY b.name, b.id, a.value""")

    POST_SQL = re.sub(r'\s+', ' ', """\
        INSERT INTO v1.project_fact_type_enums
                    (fact_type_id, value, created_by, icon_class, score)
             VALUES (%(fact_type_id)s, %(value)s, %(username)s, %(icon_class)s,
                     %(score)s)
          RETURNING id""")


class RecordRequestHandler(_RequestHandlerMixin, base.AdminCRUDRequestHandler):

    NAME = 'fact-type-enum'

    DELETE_SQL = 'DELETE FROM v1.project_fact_type_enums WHERE id=%(id)s'

    PATCH_SQL = re.sub(r'\s+', ' ', """\
        UPDATE v1.project_fact_type_enums
           SET fact_type_id=%(fact_type_id)s,
               last_modified_at=CURRENT_TIMESTAMP,
               last_modified_by=%(username)s,
               value=%(value)s,
               icon_class=%(icon_class)s,
               score=%(score)s
         WHERE id=%(id)s""")



from imbi.endpoints import base


class IndexRequestHandler(base.RequestHandler):

    NAME = 'ui-index'

    def get(self, *args, **kwargs):
        if self.request.path == '/':
            return self.redirect('/ui/')
        self.render(
            'index.html',
            javascript_url=self.application.settings.get('javascript_url'))



from imbi import errors


class PrepareFailureMixin:

    @staticmethod
    def handle_prepare_failures(action: str, failures: list) -> Exception:
        action = action.title()
        if len(failures) == 1:
            return errors.BadRequest(
                '%s failed: %s',
                action, failures[0],
                title=f'{action} failure', failures=failures)
        return errors.BadRequest(
            '%s failed with %s errors',
            action, len(failures),
            title=f'{action} failure', failures=failures)



import re

from imbi.endpoints import base, projects


class _DependencyRequestMixin:

    ID_KEY = ['project_id', 'dependency_id']
    ITEM_NAME = 'project-dependency'
    FIELDS = ['project_id', 'dependency_id']
    TTL = 300

    GET_SQL = re.sub(r'\s+', ' ', """\
        SELECT project_id, created_at, created_by, dependency_id
          FROM v1.project_dependencies
         WHERE project_id=%(project_id)s
           AND dependency_id=%(dependency_id)s""")


class CollectionRequestHandler(projects.ProjectAttributeCollectionMixin,
                               _DependencyRequestMixin,
                               base.CollectionRequestHandler):

    NAME = 'project-dependencies'

    COLLECTION_SQL = re.sub(r'\s+', ' ', """\
          SELECT project_id, created_by, dependency_id
            FROM v1.project_dependencies
           WHERE project_id=%(project_id)s
        ORDER BY dependency_id""")

    POST_SQL = re.sub(r'\s+', ' ', """\
        INSERT INTO v1.project_dependencies
                    (project_id, dependency_id, created_by)
             VALUES (%(project_id)s, %(dependency_id)s, %(username)s)
          RETURNING project_id, dependency_id""")


class RecordRequestHandler(projects.ProjectAttributeCRUDMixin,
                           _DependencyRequestMixin, base.CRUDRequestHandler):

    NAME = 'project-dependency'

    DELETE_SQL = re.sub(r'\s+', ' ', """\
        DELETE FROM v1.project_dependencies
         WHERE project_id=%(project_id)s
           AND dependency_id=%(dependency_id)s""")



import asyncio
import logging
import typing
import urllib.parse

import sprockets.mixins.http
import tornado.web
import yarl

from imbi import errors, models, version


def generate_key(project: models.Project) -> str:
    """Generate a SonarQube project key for `project`."""
    return ':'.join([project.namespace.slug.lower(), project.slug.lower()])


def generate_dashboard_link(project: models.Project,
                            sonar_settings: dict) -> typing.Union[str, None]:
    """Generate a link to the SonarQube dashboard for `project`."""
    if sonar_settings['url']:
        root = yarl.URL(sonar_settings['url'])
        return str(root.with_path('/dashboard').with_query({
            'id': generate_key(project),
        }))


class SonarQubeClient(sprockets.mixins.http.HTTPClientMixin):
    """API Client for SonarQube.

    This client uses the SonarQube HTTP API to automate aspects
    of managing projects.

    """
    enabled: typing.Union[bool, None] = None
    gitlab_alm_key: typing.Union[bool, None, str] = None

    def __init__(self, application: tornado.web.Application, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(__package__).getChild(
            'SonarQubeClient')

        self.settings = application.settings['automations']['sonarqube']
        try:
            self.admin_token = self.settings['admin_token']
            self.sonar_url = yarl.URL(self.settings['url'])
        except KeyError as error:
            if self.__class__.enabled is None:
                self.logger.warning('disabling SonarQube integration due to'
                                    ' missing configuration: %s',
                                    error.args[0])
                self.__class__.enabled = False
            self.api_root = None
            self.admin_token = None
        else:
            self.enabled = True
            self.api_root = self.sonar_url / 'api'

    async def api(self, url: typing.Union[yarl.URL, str], *,
                  method: str = 'GET', **kwargs
                  ) -> sprockets.mixins.http.HTTPResponse:
        """Make an authenticated API call."""
        if not self.enabled:
            raise RuntimeError('SonarQube integration is not enabled')
        if not isinstance(url, yarl.URL):
            url = yarl.URL(url)
        if not url.is_absolute():
            new_url = self.api_root / url.path.lstrip('/')
            url = new_url.with_query(url.query)

        request_headers = kwargs.setdefault('request_headers', {})
        request_headers['Accept'] = 'application/json'
        if kwargs.get('body', None) is not None:
            kwargs['content_type'] = 'application/x-www-form-urlencoded'
            request_headers['Content-Type'] = kwargs['content_type']
            body = urllib.parse.urlencode(kwargs.pop('body'))
            kwargs['body'] = body.encode()

        kwargs.update({
            'auth_username': self.admin_token,
            'auth_password': '',
            'user_agent': f'imbi/{version} (SonarQubeClient)',
        })
        response = await super().http_fetch(str(url), method=method, **kwargs)
        if not response.ok:
            self.logger.warning('%s %s failed: %s', method, url, response.code)
            if response.body:
                self.logger.warning('response body: %r', response.body)
        return response

    async def create_project(self, project: models.Project, *,
                             main_branch_name='main',
                             public_url: yarl.URL) -> typing.Tuple[str, str]:
        """Create a SonarQube project for `project`.

        :returns: a :class:`tuple` containing the assigned project key
            and a link to the dashboard

        """
        self.logger.info('creating SonarQube project for %s', project.slug)
        response = await self.api('/projects/create', method='POST', body={
            'name': project.name,
            'project': generate_key(project),
        })
        if not response.ok:
            raise errors.InternalServerError(
                'failed to create project %s: %s', project.name, response.code,
                title='SonarQube API Failure', sonar_response=response.body)
        project_key = response.body['project']['key']
        dashboard_url = self.sonar_url.with_path('/dashboard').with_query({
            'id': project_key,
        })

        await asyncio.gather(
            self._fix_main_branch(project_key, main_branch_name),
            self._enable_pr_decoration(project_key, project.gitlab_project_id),
            self._add_link_to_sonar(project_key, 'Imbi Project', public_url),
        )
        return project_key, str(dashboard_url)

    async def _fix_main_branch(self, project_key: str, main_branch_name: str):
        """Sonar assumes that the "main" branch is named master.

        This will cause main branches with any other name to be
        treated as one-off branches.  You can *manually* fix this
        by deleting the branch that was analyzed and renaming the
        "master" branch.  Instead of requiring this for projects,
        this method ensures that the "master" branch in SonarQube
        matches the default branch in the SCM.

        """
        response = await self.api(
            yarl.URL('/project_branches/list').with_query({
                'project': project_key,
            }))
        if not response.ok:
            self.logger.error('failed to list project branches for %s: %s',
                              project_key, response.code)
            return

        self.logger.debug('making sure that main branch is %s for %s',
                          main_branch_name, project_key)
        main_branches = [branch
                         for branch in response.body['branches']
                         if branch['isMain']]
        if main_branches and main_branches[0]['name'] != main_branch_name:
            branch = main_branches[0]
            self.logger.info('resetting main branch from %s to %s for %s',
                             branch['name'], main_branch_name, project_key)
            response = await self.api(
                '/project_branches/rename', method='POST', body={
                    'project': project_key,
                    'name': main_branch_name,
                })
            if not response.ok:
                self.logger.error(
                    'failed to rename main branch %s for %s: %s',
                    branch['name'], project_key, response.code)

    async def _enable_pr_decoration(self, project_key: str,
                                    gitlab_project_id: int):
        """Enable GitLab MR decoration if it is available."""
        alm_enabled = await self._is_gitlab_alm_available()
        if not alm_enabled:
            return

        self.logger.debug('checking for PR decoration on %s', project_key)
        response = await self.api(
            yarl.URL('/alm_settings/get_binding').with_query({
                'project': project_key,
            })
        )
        if response.code == 404:  # not configured or doesn't exist
            response = await self.api(
                yarl.URL('/alm_settings/set_gitlab_binding'),
                method='POST',
                body={
                    'almSetting': self.gitlab_alm_key,
                    'project': project_key,
                    'repository': gitlab_project_id,
                }
            )
            if not response.ok:
                self.logger.error(
                    'failed to enable PR decoration for %s: %s',
                    project_key, response.code)
        elif not response.ok:
            self.logger.error(
                'failed to check for GitLab integration for %s: %s',
                project_key, response.code)

    async def _is_gitlab_alm_available(self) -> bool:
        """Check if the SonarQube server has the GitLab ALM configured.

        The result of this method is "remembered" by setting the
        `gitlab_alm_key` class attribute to something other than
        :data:`None`.  It will be set to :data:`False` when we
        determine that the GitLab ALM is not configured or to the
        configured "key" value.

        """
        if self.gitlab_alm_key is None:
            response = await self.api('/alm_settings/list_definitions')
            if response.ok:
                self.__class__.gitlab_alm_key = False
                for connection in response.body.get('gitlab', []):
                    if connection['key'] == 'gitlab':
                        self.logger.info('found GitLab ALM %s',
                                         connection['key'])
                        self.__class__.gitlab_alm_key = connection['key']
                        break
                else:
                    self.logger.warning(
                        'GitLab ALM not found in %r, disabling MR decoration',
                        response.body)
            else:
                self.logger.warning(
                    'failed to list alm_settings for sonar: %s',
                    response.code)
                return False  # don't cache failures for now

        return bool(self.gitlab_alm_key)

    async def _add_link_to_sonar(self, project_key: str, name: str,
                                 url: yarl.URL):
        """Add a named link to the SonarQube dashboard."""
        self.logger.debug('adding link to Imbi project for %s', project_key)
        response = await self.api(
            yarl.URL('/project_links/create'),
            method='POST',
            body={'name': name, 'projectKey': project_key, 'url': str(url)})
        if not response.ok:
            self.logger.error('failed to set the Imbi project link for %s: %s',
                              project_key, response.code)



from __future__ import annotations

from imbi import models
from imbi.automations import base
from imbi.clients import sentry


class SentryCreateProjectAutomation(base.Automation):

    def __init__(self, application, project_id, current_user, db) -> None:
        super().__init__(application, current_user, db)
        self.client = sentry.SentryClient(application)
        self.imbi_project_id = project_id
        self.project: models.Project | None = None
        self.settings = self.automation_settings['sentry']

    async def prepare(self) -> list[str]:
        if not self.client.enabled:
            self._add_error('Sentry integration is not enabled')
        else:
            project = await self._get_project(self.imbi_project_id)
            if project is None:
                self._add_error('Project id {} does not exist',
                                self.imbi_project_id)
            elif not project.namespace.sentry_team_slug:
                self._add_error(
                    'Namespace {} does not have a sentry slug, not creating'
                    ' sentry project for project {}', project.namespace.slug,
                    self.imbi_project_id)
            else:
                self.project = project
        return self.errors

    async def run(self) -> sentry.ProjectInfo:
        project = await self.client.create_project(
            self.project.namespace.sentry_team_slug, self.project.name)
        try:
            await self.db.execute(
                'UPDATE v1.projects'
                '   SET sentry_project_slug = %(slug)s,'
                '       last_modified_at = CURRENT_TIMESTAMP,'
                '       last_modified_by = %(username)s'
                ' WHERE id = %(imbi_project_id)s', {
                    'imbi_project_id': self.imbi_project_id,
                    'slug': project.slug,
                    'username': self.user.username,
                }
            )
            if self.settings.get('project_link_type_id'):
                await self.db.execute(
                    '  INSERT INTO v1.project_links(project_id, link_type_id,'
                    '                               created_by, url)'
                    '       VALUES (%(imbi_project_id)s,'
                    '               %(project_link_type_id)s, %(username)s,'
                    '               %(url)s) '
                    '  ON CONFLICT '
                    'ON CONSTRAINT project_links_pkey'
                    '    DO UPDATE'
                    '          SET last_modified_at = CURRENT_TIMESTAMP,'
                    '              last_modified_by = %(username)s,'
                    '              url = %(url)s', {
                        'imbi_project_id': self.imbi_project_id,
                        'project_link_type_id':
                            self.settings['project_link_type_id'],
                        'url': project.link,
                        'username': self.user.username,
                    }
                )
            for name, value in project.keys.items():
                await self.db.execute(
                    '  INSERT INTO v1.project_secrets(project_id, name, value,'
                    '                                 created_by)'
                    '       VALUES (%(imbi_project_id)s, %(key_name)s,'
                    '               %(key_value)s, %(username)s)'
                    '  ON CONFLICT '
                    'ON CONSTRAINT project_secrets_pkey'
                    '    DO UPDATE'
                    '          SET value = %(key_value)s,'
                    '              last_modified_at = CURRENT_TIMESTAMP,'
                    '              last_modified_by = %(username)s', {
                        'imbi_project_id': self.imbi_project_id,
                        'key_name': f'sentry_{name.lower()}',
                        'key_value': self.application.encrypt_value(value),
                        'username': self.user.username,
                    }
                )

        # Using a wide catch of Exception here to ensure that we remove
        # the sentry project if we fail to save it for any reason.  This
        # prevents orphaning sentry projects when we have a DB problem.
        except Exception as error:
            self.logger.error('removing sentry project %s due to error: %s',
                              project.slug, error)
            await self.client.remove_project(project.slug)
            raise

        return project



"""
System Reports

"""
from tornado import web

from . import automations, authentication, groups, index, settings, user

IndexRequestHandler = index.IndexRequestHandler

URLS = automations.URLS + [
    web.url(r'^/ui/', index.IndexRequestHandler),
    web.url(r'^/ui/login$', authentication.LoginRequestHandler),
    web.url(r'^/ui/logout$', authentication.LogoutRequestHandler),
    web.url(r'^/ui/settings$', settings.RequestHandler),
    web.url(r'^/ui/user$', user.UserRequestHandler),
    web.url(r'^/ui/.*$', index.IndexRequestHandler)
]



"""
API Endpoint for returning UI Settings

"""
import yarl

from imbi import errors
from imbi.automations import sonarqube
from imbi.endpoints import base
from imbi.endpoints.ui.automations import mixins


class CreationRequestHandler(mixins.PrepareFailureMixin,
                             base.AuthenticatedRequestHandler):

    NAME = 'ui-sonarqube-create'

    async def post(self):
        request = self.get_request_body()
        try:
            imbi_project_id = request['project_id']
        except KeyError as error:
            raise errors.BadRequest('request missing required field %s',
                                    error.args[0])

        if self.settings.get('frontend_url'):
            public_url = yarl.URL(self.settings['frontend_url'])
        else:
            public_url = yarl.URL.build(scheme=self.request.protocol,
                                        host=self.request.host)
        public_url = public_url.with_path(
            self.reverse_url('project', imbi_project_id))

        async with self.postgres_transaction() as transaction:
            automation = sonarqube.SonarCreateProject(
                self.application, imbi_project_id, public_url,
                await self.get_current_user(), transaction)
            failures = await automation.prepare()
            if failures:
                raise self.handle_prepare_failures('Create SonarQube Project',
                                                   failures)
            project_info = await automation.run()
        self.send_response(project_info)



from tornado import web


class StaticFileHandler(web.StaticFileHandler):

    def set_default_headers(self) -> None:
        """Override the default headers, setting the Server response header"""
        super().set_default_headers()
        self.set_header('Server', self.settings['server_header'])
        self.set_header('Access-Control-Allow-Origin', '*')



"""
Application Status View

"""
import platform

import distro
from sprockets.mixins import mediatype
from tornado import web

from imbi import version

MAINTENANCE = 'maintenance'
OK = 'ok'


class RequestHandler(mediatype.ContentMixin,
                     web.RequestHandler):
    """Returns the current status"""
    NAME = 'status'

    SYSTEM = {
        'language': {
            'name': 'python',
            'implementation': platform.python_implementation(),
            'version': platform.python_version()
        },
        'os': {
            'name': distro.name(),
            'version': distro.version()
        }
    }

    async def get(self):
        """Tornado RequestHandler GET request endpoint for reporting status"""
        self.set_status(self._status_response_code())
        self.send_response({
            'started_at': self.application.started_at_str,
            'status': OK if self.application.ready_to_serve else MAINTENANCE,
            'system': self.SYSTEM,
            'version': version})

    def set_default_headers(self) -> None:
        """Override the default headers, setting the Server response header"""
        super().set_default_headers()
        self.set_header('Server', self.settings['server_header'])

    def _status_response_code(self):
        """Return the status code for the request based upon the application
        status.
        :rtype: int
        """
        if self.application.ready_to_serve:
            return 200
        return 503



"""
Imbi CLI application

"""
import argparse
import asyncio
import base64
import binascii
import logging
import pathlib
import sys
import typing
from logging import config as logging_config

import yaml
from sprockets import http
from tornado import ioloop

from imbi import app, pkgfiles, version
from imbi.endpoints import static
from imbi.opensearch import project

LOGGER = logging.getLogger(__name__)

DEFAULT_LOG_CONFIG = {
    'version': 1,
    'formatters': {
        'verbose': {
            'format':
                '%(levelname) -10s %(asctime)s %(process)-6d '
                '%(name) -20s %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S'
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'verbose'
        }
    },
    'loggers': {
        'imbi': {
            'level': 'INFO'
        },
        'opensearch': {
            'level': 'INFO'
        },
        'tornado': {
            'level': 'INFO'
        }
    },
    'root': {
        'level': 'WARNING',
        'handlers': ['console']
    },
    'disable_existing_loggers': True,
    'incremental': False
}


def run() -> None:
    args = _parse_cli_args()
    config = load_configuration(args.config[0], args.debug)
    if args.initialize:
        return asyncio.run(initialize(config[0], config[1], args.build))
    http.run(app.Application, *config)


async def initialize(settings: dict, logging_settings: dict, build: bool):
    logging_config.dictConfig(logging_settings)
    loop = ioloop.IOLoop.current()

    """Initialize the OpenSearch Indexes"""
    application = app.Application(**settings)
    await application._postgres_on_start(application, loop)
    await application.on_start(application, loop)

    # Initialize the Projects Index
    await project.initialize(application, build)

    LOGGER.info('Initialization is complete')
    await application._postgres_shutdown(loop)
    await application.on_shutdown()


def load_configuration(config: str, debug: bool) -> typing.Tuple[dict, dict]:
    """Load the configuration file and apply the default settings"""
    config_file = pathlib.Path(config)
    if not config_file.exists():
        sys.stderr.write(
            'Configuration file {} not found\n'.format(config))
        sys.exit(1)

    with config_file.open('r') as handle:
        try:
            config = yaml.safe_load(handle)
        except yaml.YAMLError as error:
            sys.stderr.write(
                'Failed to load configuration file: {}\n'.format(error))
            sys.exit(1)
        else:
            if not isinstance(config, dict):
                sys.stderr.write(
                    'Configuration file {} is not a YAML mapping\n'.format(
                        config_file.name))
                sys.exit(1)

    log_config = config.get('logging', DEFAULT_LOG_CONFIG)

    automations = config.get('automations', {})
    automations_grafana = automations.get('grafana', {})
    if automations_grafana.get('url') \
            and automations_grafana.get('admin_token'):
        automations_grafana['enabled'] = True
    automations_sonar = automations.get('sonarqube', {})
    if automations_sonar.get('url') and automations_sonar.get('admin_token'):
        automations_sonar['enabled'] = True
    footer_link = config.get('footer_link', {})
    http_settings = config.get('http', {})
    ldap = config.get('ldap', {})
    postgres = config.get('postgres', {})
    sentry = config.get('sentry', {})
    session = config.get('session', {})
    stats = config.get('stats', {})

    automations_gitlab = automations.get('gitlab', {})
    automations_gitlab.setdefault(
        'enabled',
        automations_gitlab.get('project_link_type_id') is not None)

    automations_sentry = automations.get('sentry', {})
    automations_sentry.setdefault(
        'enabled',
        (automations_sentry.get('auth_token') is not None and
         automations_sentry.get('organization') is not None))
    automations_sentry.setdefault('url', 'https://sentry.io/')

    module_path = pathlib.Path(sys.modules['imbi'].__file__).parent

    # Allow encryption key to be encoded as a Base-64 string.
    # If it isn't, then use it as-is.
    try:
        encoded_key = config['encryption_key']
    except KeyError:
        encryption_key = b'some thirty-two character secret'
    else:
        try:
            encryption_key = base64.b64decode(encoded_key.encode())
        except binascii.Error:
            encryption_key = encoded_key.encode()

    settings = {
        'automations': {
            'gitlab': automations_gitlab,
            'grafana': {
                'enabled': automations_grafana.get('enabled', False),
                'project_link_type_id':
                    automations_grafana.get('project_link_type_id')
            },
            # see https://pycqa.github.io/isort/reference/isort/settings.html
            'isort': automations.get('isort', {}),
            'sentry': automations_sentry,
            'sonarqube': {
                'enabled': automations_sonar.get('enabled', False),
                'admin_token': automations_sonar.get('admin_token'),
                'project_link_type_id':
                    automations_sonar.get('project_link_type_id'),
                'url': automations_sonar.get('url')
            },
            # see https://github.com/google/yapf#knobs
            'yapf': automations.get('yapf', {}),
        },
        'canonical_server_name': http_settings['canonical_server_name'],
        'compress_response': http_settings.get('compress_response', True),
        'cookie_secret': http_settings.get('cookie_secret', 'imbi'),
        'cors': config.get('cors', None),
        'debug': debug,
        'encryption_key': encryption_key,
        'footer_link': {
            'icon': footer_link.get('icon', ''),
            'text': footer_link.get('text', ''),
            'url': footer_link.get('url', '')
        },
        'frontend_url': config.get('frontend_url', None),
        'javascript_url': config.get('javascript_url', None),
        'ldap': {
            'enabled': ldap.get('enabled'),
            'host': ldap.get('host', 'localhost'),
            'port': ldap.get('port', 636),
            'ssl': ldap.get('ssl', False),
            'pool_size': ldap.get('pool_size', 5),
            'group_member_attr': ldap.get('group_member_dn', 'member'),
            'group_object_type': ldap.get('group_object_type', 'groupOfNames'),
            'groups_dn': ldap.get('groups_dn'),
            'user_object_type': ldap.get('user_object_type', 'inetOrgPerson'),
            'username': ldap.get('username', 'uid'),
            'users_dn': ldap.get('users_dn')
        },
        'number_of_procs': http_settings.get('processes', 2),
        'opensearch': config.get('opensearch', {}),
        'permissions': [],
        'port': http_settings.get('port', 8000),
        'postgres_url': postgres.get('url'),
        'postgres_max_pool_size': postgres.get('max_pool_size'),
        'postgres_min_pool_size': postgres.get('min_pool_size'),
        'postgres_connection_timeout': postgres.get('connection_timeout'),
        'postgres_connection_ttl': postgres.get('connection_ttl'),
        'postgres_query_timeout': postgres.get('query_timeout'),
        'project_url_template': config.get('project_url_template', None),
        'sentry_backend_dsn': sentry.get('backend_dsn'),
        'sentry_ui_dsn': sentry.get('ui_dsn', 'false'),
        'server_header': 'imbi/{}'.format(version),
        'session_duration': int(session.get('duration', '7')),
        'session_pool_size': session.get('pool_size', 10),
        'session_redis_url': session.get(
            'redis_url', 'redis://localhost:6379/0'),
        'static_handler_class': static.StaticFileHandler,
        'static_path': module_path / 'static',
        'stats_pool_size': stats.get('pool_size', 10),
        'stats_redis_url': stats.get(
            'redis_url', 'redis://localhost:6379/1'),
        'template_loader': pkgfiles.TemplateLoader(debug=debug),
        'template_path': module_path / 'templates',
        'version': version,
        'xheaders': http_settings.get('xheaders', True),
        'xsrf_cookies': False,
    }

    return {k: v for k, v in settings.items() if v is not None}, log_config


def _parse_cli_args() -> argparse.Namespace:
    """Create the CLI parser and parse the CLI arguments"""
    parser = argparse.ArgumentParser(
        'Imbi',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '--debug', action='store_true', help='Enable debug mode')
    parser.add_argument(
        '--initialize', action='store_true', help='Initialize Search Indexes')
    parser.add_argument(
        '--build', action='store_true',
        help='Build the search index when initializing')
    parser.add_argument('-V', '--version', action='version', version=version)
    parser.add_argument(
        'config', metavar='CONFIG FILE', nargs=1,
        help='Configuration File')
    return parser.parse_args()



import typing

from imbi import errors, oauth2
from imbi.automations import gitlab as automations
from imbi.clients import gitlab
from imbi.endpoints import base
from imbi.endpoints.ui.automations import mixins


class GitLabAutomationRequestHandler(mixins.PrepareFailureMixin,
                                     base.AuthenticatedRequestHandler):

    NAME = 'ui-gitlab'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gitlab: typing.Optional[gitlab.GitLabClient] = None

    async def prepare(self):
        await super().prepare()
        if not self._finished:
            integration = await oauth2.OAuth2Integration.by_name(
                self.application, 'gitlab')
            if not integration:
                raise errors.IntegrationNotFound('gitlab')

            user = await self.get_current_user()  # never None
            tokens = await integration.get_user_tokens(user)
            if not tokens:
                raise errors.Forbidden(
                    '%r has no GitLab tokens', user,
                    title='GitLab Not Connected')
            self.gitlab = gitlab.GitLabClient(tokens[0], self.application)


class CreationRequestHandler(GitLabAutomationRequestHandler,
                             base.ValidatingRequestHandler):

    NAME = 'ui-gitlab-create'

    async def post(self):
        request = self.get_request_body()
        project_id = int(request['project_id'])

        async with self.postgres_transaction() as transaction:
            automation = automations.GitLabCreateProjectAutomation(
                self.application, project_id, await self.get_current_user(),
                transaction)
            failures = await automation.prepare()
            if failures:
                raise self.handle_prepare_failures('Create Project', failures)
            gitlab_info = await automation.run()

        self.send_response({
            'gitlab_project_id': gitlab_info.id,
            'gitlab_project_url': gitlab_info.links.self,
        })


class InitialCommitRequestHandler(GitLabAutomationRequestHandler):

    NAME = 'ui-gitlab-commit'

    async def post(self):
        request = self.get_request_body()
        try:
            imbi_project_id = request['project_id']
            cookie_cutter = request['cookie_cutter']
        except KeyError as error:
            raise errors.BadRequest(
                'Error creating initial commit: missing required field %s',
                error.args[0])
        except TypeError as error:
            raise errors.BadRequest('Error creating initial commit: %s', error)

        async with self.postgres_transaction() as transaction:
            automation = automations.GitLabInitialCommitAutomation(
                self.application, imbi_project_id, cookie_cutter,
                await self.get_current_user(), transaction)
            failures = await automation.prepare()
            if failures:
                raise self.handle_prepare_failures('Create Initial Commit',
                                                   failures)
            try:
                commit_info = await automation.run()
            except automations.CookieCutterError as error:
                raise errors.InternalServerError(
                    'Error applying cookiecutter: %s', error)
            except automations.InitialCommitError as error:
                raise errors.InternalServerError(
                    'Error creating initial commit: %s', error)

        self.send_response(commit_info)



"""
LDAP Client
===========

"""
import dataclasses
import logging
import typing
from concurrent import futures

import ldap3
from ldap3.core import exceptions

LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass
class Settings:
    enabled: bool
    host: str
    port: int
    ssl: bool
    group_member_attr: str
    group_object_type: str
    groups_dn: str
    user_object_type: str
    users_dn: str
    username: str
    pool_size: int


class Client:
    """LDAP Client used for authenticating users. Expects the following
    configuration in the settings dict:

    - ``enabled``: Enables / Disables LDAP lookup
    - ``host``: The hostname of the server to connect to
    - ``port``: The port to connect on
    - ``ssl``: Indicates whether SSL is enabled for connecting
    - ``group_object_type``: The object type to use for groups
    - ``groups_dn``: The base DN to use for group searching
    - ``group_member_attr``: The group membership attribute used in searching
    - ``user_object_type``: The object type to use for users
    - ``users_dn``: The base DN to use for user searching
    - ``username``: The username attribute used in user searching
    - ``pool_size``: The size to allocate for the ThreadPoolExecutor

    """
    def __init__(self, settings: dict):
        self._settings = Settings(**settings)
        self._executor = futures.ThreadPoolExecutor(
            max_workers=self._settings.pool_size)
        self._server = ldap3.Server(
            self._settings.host, self._settings.port, self._settings.ssl)

    def __exit__(self, _exc_type, _exc_val, _exc_tb) -> None:
        """Shutdown the executor when the class is destroyed"""
        self._executor.shutdown()

    @property
    def is_enabled(self) -> bool:
        """Returns a boolean indicating if LDAP is enabled"""
        return self._settings.enabled

    async def connect(self, username: str, password: str) -> ldap3.Connection:
        """Connect to the LDAP server with the specified username
        and password. Returns a LDAP connection object for use with
        other methods.

        :param username: The username to connect with
        :param password: The password to use
        :raises: RuntimeError

        """
        if not self._settings.enabled:
            raise RuntimeError('LDAP not enabled')
        future = self._executor.submit(self._connect, username, password)
        conn = future.result()
        LOGGER.debug('Connected to %s', conn)
        return conn

    async def attributes(self, conn: ldap3.Connection, dn: str) -> dict:
        """Return the attributes for the specified dn on the given
        connection.

        :param conn: The connection to use
        :param dn: The DN of the user to return the attributes for
        :raises: RuntimeError

        """
        if not self._settings.enabled:
            raise RuntimeError('LDAP not enabled')
        future = self._executor.submit(self._attributes, conn, dn)
        return future.result()

    async def groups(self, conn: ldap3.Connection, dn: str) -> list:
        """Return the names of the groups for the specified username.

        :param conn: The connection to use
        :param dn: The DN of the user to return the groups for
        :raises: RuntimeError

        """
        if not self._settings.enabled:
            raise RuntimeError('LDAP not enabled')
        future = self._executor.submit(self._groups, conn, dn)
        return future.result()

    def _connect(self, username: str, password: str) -> ldap3.Connection:
        """Connect to the LDAP server with the specified username and password.

        :param username: The username to connect as
        :param password: The password to use
        :raises: RuntimeError

        """
        if not self._settings.enabled:
            raise RuntimeError('LDAP not enabled')
        dn = ','.join(['{}={}'.format(self._settings.username, username),
                       self._settings.users_dn])
        LOGGER.debug(
            'Connecting to %s://%s:%s as %s',
            'ldaps' if self._settings.ssl else 'ldap',
            self._settings.host, self._settings.port, dn)
        try:
            return ldap3.Connection(
                self._server, dn, password, auto_bind=True)
        except exceptions.LDAPException as error:
            LOGGER.warning('Authentication error for %s: %s', username, error)

    def _attributes(self, conn: ldap3.Connection, dn: str) -> dict:
        """Return the attributes for the specified username.

        :param conn: The connection to use
        :param dn: The DN of the user to return the groups for

        """
        conn.search(
            dn, '(objectClass={})'.format(self._settings.user_object_type),
            attributes=[ldap3.ALL_ATTRIBUTES,
                        ldap3.ALL_OPERATIONAL_ATTRIBUTES])
        return {k: v[0] if isinstance(v, list) and len(v) == 1 else v
                for k, v in conn.response[0]['attributes'].items()}

    def _groups(self, conn: ldap3.Connection, dn: str) -> typing.List[str]:
        """Return the groups for the specified username.

        :param conn: The connection to use
        :param dn: The DN of the user to return the groups for

        """
        query = '(&(objectClass={})({}={}))'.format(
            self._settings.group_object_type,
            self._settings.group_member_attr, dn)
        LOGGER.debug('User groups query: %r', query)
        conn.search(self._settings.groups_dn, query)
        groups = set({})
        for entry in conn.response:
            groups.add(str(entry['dn']))
        return sorted(groups)



import re

from imbi.endpoints import base


class _RequestHandlerMixin:

    ID_KEY = 'id'
    FIELDS = ['id', 'link_type', 'icon_class']
    DEFAULTS = {'icon_class': 'fas fa-link'}

    GET_SQL = re.sub(r'\s+', ' ', """\
        SELECT id, created_at, created_by, last_modified_at, last_modified_by,
               link_type, icon_class
          FROM v1.project_link_types
         WHERE id=%(id)s""")


class CollectionRequestHandler(_RequestHandlerMixin,
                               base.CollectionRequestHandler):

    NAME = 'project-link-types'
    ITEM_NAME = 'project-link-type'

    COLLECTION_SQL = re.sub(r'\s+', ' ', """\
        SELECT id, link_type, icon_class
          FROM v1.project_link_types
         ORDER BY link_type ASC""")

    POST_SQL = re.sub(r'\s+', ' ', """\
        INSERT INTO v1.project_link_types
                    (link_type, created_by, icon_class)
             VALUES (%(link_type)s, %(username)s, %(icon_class)s)
          RETURNING id""")


class RecordRequestHandler(_RequestHandlerMixin, base.AdminCRUDRequestHandler):

    NAME = 'project-link-type'

    DELETE_SQL = 'DELETE FROM v1.project_link_types WHERE id=%(id)s'

    PATCH_SQL = re.sub(r'\s+', ' ', """\
        UPDATE v1.project_link_types
           SET link_type=%(link_type)s,
               last_modified_at=CURRENT_TIMESTAMP,
               last_modified_by=%(username)s,
               icon_class=%(icon_class)s
         WHERE id=%(id)s""")



from __future__ import annotations

import re
import typing

import psycopg2.errors

from imbi import common, errors
from imbi.endpoints import base
from imbi.opensearch import project


class CollectionRequestHandler(project.RequestHandlerMixin,
                               base.CollectionRequestHandler):

    NAME = 'project-fact-types'
    ID_KEY = 'project_id'

    COLLECTION_SQL = re.sub(r'\s+', ' ', """\
        WITH project_type_id AS (SELECT project_type_id AS id
                                   FROM v1.projects
                                  WHERE id = %(project_id)s)
        SELECT a.id AS fact_type_id,
               a.name,
               b.recorded_at,
               b.recorded_by,
               b.value,
               a.data_type,
               a.ui_options,
               CASE WHEN b.value IS NULL THEN 0
                    ELSE CASE WHEN a.fact_type = 'enum' THEN (
                                          SELECT score::NUMERIC(9,2)
                                            FROM v1.project_fact_type_enums
                                           WHERE fact_type_id = b.fact_type_id
                                             AND value = b.value)
                              WHEN a.fact_type = 'range' THEN (
                                          SELECT score::NUMERIC(9,2)
                                            FROM v1.project_fact_type_ranges
                                           WHERE fact_type_id = b.fact_type_id
                                             AND b.value::NUMERIC(9,2)
                                         BETWEEN min_value AND max_value)
                              WHEN a.data_type = 'boolean' AND b.value = 'true'
                              THEN 100
                              ELSE 0
                          END
                END AS score,
               a.weight
          FROM v1.project_fact_types AS a
     LEFT JOIN v1.project_facts AS b
            ON b.fact_type_id = a.id
           AND b.project_id = %(project_id)s
         WHERE (SELECT id FROM project_type_id) = ANY(a.project_type_ids)
        ORDER BY a.name""")

    POST_SQL = re.sub(r'\s+', ' ', """\
        INSERT INTO v1.project_facts
                    (project_id, fact_type_id, recorded_at, recorded_by, value)
             VALUES (%(project_id)s, %(fact_type_id)s, CURRENT_TIMESTAMP,
                     %(username)s, %(value)s)
        ON CONFLICT (project_id, fact_type_id)
          DO UPDATE SET recorded_at = CURRENT_TIMESTAMP,
                        recorded_by = %(username)s,
                        value = %(value)s""")

    async def get(self, *args, **kwargs):
        result = await self.postgres_execute(
            self.COLLECTION_SQL, self._get_query_kwargs(kwargs),
            'get-{}'.format(self.NAME))
        self.send_response(common.coerce_project_fact_values(result.rows))

    async def post(self, *args, **kwargs):
        facts = await self._coerce_facts(self.get_request_body())
        for fact in facts:
            fact.update({
                'project_id': kwargs['project_id'],
                'username': self._current_user.username
            })
            await self.postgres_execute(
                self.POST_SQL, fact, 'post-{}'.format(self.NAME))
        await self.index_document(kwargs['project_id'])
        self.set_status(204)

    async def _coerce_facts(self, facts: list[dict]) -> list[dict]:
        """Take a list of project facts and coerce their values.

        This method looks up the data types and then calls
        :func:`common.coerce_project_fact_values` to perform
        the coercion appropriately.

        """
        result = await self.postgres_execute(
            'SELECT id, data_type'
            '  FROM v1.project_fact_types'
            ' WHERE id IN %(fact_type_ids)s',
            {'fact_type_ids': tuple(fact['fact_type_id'] for fact in facts)},
        )
        fact_type_map = {row['id']: row['data_type'] for row in result.rows}
        for fact in facts:
            try:
                fact_type = fact_type_map[fact['fact_type_id']]
            except KeyError:
                raise errors.BadRequest('%s is not a known fact type id',
                                        fact['fact_type_id'])
            else:
                fact['data_type'] = fact_type

        try:
            return common.coerce_project_fact_values(facts)
        except ValueError as error:
            raise errors.BadRequest('invalid fact value: %s', error)

    def on_postgres_error(self,
                          metric_name: str,
                          exc: Exception) -> typing.Optional[Exception]:
        """Invoked when an error occurs when executing a query

        If `tornado-problem-details` is available,
        :exc:`problemdetails.Problem` will be raised instead of
        :exc:`tornado.web.HTTPError`.

        Override for different error handling behaviors.

        Return an exception if you would like for it to be raised, or swallow
        it here.

        """
        if isinstance(exc, psycopg2.errors.lookup('P0001')):
            return errors.BadRequest(str(exc).split('\n')[0])
        super().on_postgres_error(metric_name, exc)



"""
System Reports

"""
from tornado import web

from . import namespace_kpis, namespace_shs_history, system_shs_history

URLS = [
    web.url(r'/reports/namespace-kpis',
            namespace_kpis.RequestHandler),
    web.url(r'/reports/namespace-shs-history',
            namespace_shs_history.RequestHandler),
    web.url(r'/reports/system-shs-history',
            system_shs_history.RequestHandler),

]



from __future__ import annotations

import copy
import logging

import yarl
from tornado import web

LOGGER = logging.getLogger(__name__)


class Unspecified:
    """Sentinel value for optional parameters"""
    pass


class Origins:
    """Collection of CORS origins

    This acts like a set of URLs with a special property to
    enable any URL to match without being present.  This is
    used to determine whether CORS should be enabled for
    requests from a specific Origin.

    By default, any origin is acceptable.  If the application
    adds a specific origin value, then the "any" flag is
    disabled since that is usually what is desired.  The "add"
    operation enforces that only absolute URLs are acceptable
    as origins -- this explicitly means that ``null`` is not
    valid as an acceptable origin. (:rfc:`6454`).

    """
    def __init__(self, *, allow_any: bool = True):
        self.allow_any: bool = allow_any
        self._origins: set[yarl.URL] = set()

    def add(self, item: str) -> None:
        """Add a new URL to the set of origins

        If `item` is not an absolute URL, then a :exc:`ValueError`
        will be raised.

        """
        url = yarl.URL(item)
        if not url.is_absolute():
            raise ValueError(f'{item!r} is not a valid URL')

        self.allow_any = False
        self._origins.add(url)

    def __str__(self) -> str:
        return '<{} allow_any:{} origins:{}>'.format(
            self.__class__.__name__,
            self.allow_any,
            ','.join(str(s) for s in self._origins))

    def __copy__(self) -> Origins:
        return self.__deepcopy__()

    def __deepcopy__(self, *args, **kwargs) -> Origins:
        clone = self.__class__(allow_any=self.allow_any)
        clone._origins.update(self._origins)
        return clone

    def __contains__(self, item: str) -> bool:
        if self.allow_any:
            return True
        return yarl.URL(item) in self._origins


class CORSConfig:
    """Application-level CORS configuration defaults

    :param allow_any_origin: should "any" origin be allowed
        to participate in resource sharing?
    :param allow_credentials: should requests including
        an Authorization header or Cookies be sent to the server?
        Also, should responses with Set-Cookie be returned?
    :param exposed_headers: the list of response headers that
        should be exposed to the JS layer
    :param max_age: the maximum number of seconds that the JS
        layer may cache the CORS response

    """

    def __init__(
            self,
            *,
            allow_any_origin: bool = True,
            allow_credentials: bool = True,
            allow_methods: set[str] | None = None,
            exposed_headers: set[str] | None = None,
            max_age: int = 5,
    ):
        # allow_credentials is True by default to ensure that any
        # cookies we set will be returned to the JS client
        self.allow_credentials = allow_credentials

        self.allowed_methods: set[str] = set()
        if allow_methods:
            self.allowed_methods.update(allow_methods)

        self.allowed_origins = Origins(allow_any=allow_any_origin)

        self.exposed_headers: set[str] = {'Cache-Control', 'Date',
                                          'Last-Modified', 'Link'}
        if exposed_headers:
            self.exposed_headers.update(exposed_headers)

        self.max_age = max_age

    def __str__(self) -> str:
        return (('<{} allow_credentials:{} max_age:{} allowed_methods:{!r}'
                 ' exposed_headers:{!r} origins:{}>').format(
            self.__class__.__name__,
            self.allow_credentials,
            self.max_age,
            self.allowed_methods,
            self.exposed_headers,
            self.allowed_origins,
        ))

    def __copy__(self) -> CORSConfig:
        clone = CORSConfig(
            allow_credentials=self.allow_credentials,
            allow_methods=self.allowed_methods,
            exposed_headers=self.exposed_headers,
            max_age=self.max_age)
        clone.allowed_origins = copy.deepcopy(self.allowed_origins)
        return clone

    def __deepcopy__(self, *args, **kwargs) -> CORSConfig:
        return self.__copy__()

    def update(self, *,
               allow_any_origin: bool | Unspecified = Unspecified(),
               allow_credentials: bool | Unspecified = Unspecified(),
               allow_methods: set[str] | Unspecified = Unspecified(),
               allow_origins: set[str] | Unspecified = Unspecified(),
               exposed_headers: set[str] | Unspecified = Unspecified(),
               max_age: int | Unspecified = Unspecified(),
               ) -> None:
        """Update selected attributes"""
        if not isinstance(allow_any_origin, Unspecified):
            self.allowed_origins.allow_any = allow_any_origin
        if not isinstance(allow_credentials, Unspecified):
            self.allow_credentials = allow_credentials
        if not isinstance(allow_methods, Unspecified):
            self.allowed_methods.update(allow_methods)
        if not isinstance(allow_origins, Unspecified):
            for origin in allow_origins:
                self.allowed_origins.add(origin)
        if not isinstance(exposed_headers, Unspecified):
            self.exposed_headers.update(exposed_headers)
        if not isinstance(max_age, Unspecified):
            self.max_age = max_age


class CORSProcessor:
    """CORS Request processor

    An instance of this is created for each request using the
    default configuration as the `config` parameter along with
    optional overrides.  It is instantiated with each handler
    instance, then :meth:`process_request` is called inside
    :meth:`tornado.web.RequestHandler.prepare`.  "Processing"
    the request results in setting the response headers based
    on the CORS protocol.

    The separation between ``process_request`` and ``set_headers``
    is necessary since the request handler logic resets the header
    set when an exception occurs.  The ``CORSMixin`` accounts for
    this by calling :meth:`set_headers` in ``set_default_headers``
    which is called after the request handler resets the headers.
    However, it is also called during instance initialization which
    the mixin also accounts for.  You should rarely need to call
    `set_headers` directly.

    """
    def __init__(self, config: CORSConfig, **overrides) -> None:
        self.config = copy.deepcopy(config)
        self.config.update(**overrides)
        self.requested_origin: str | None = None
        self.is_preflight = None

    @property
    def ok(self) -> bool:
        """Is the request Origin acceptable?"""
        return self.requested_origin is not None

    def process_request(self, handler: web.RequestHandler) -> None:
        """Determine the CORS result for a request

        This is the main logic for CORS handling and should be
        called from your request handler's ``prepare`` method.

        """
        request = handler.request
        if request.method == 'OPTIONS':
            try:
                request.headers['Access-Control-Request-Method']
                origin = request.headers['Origin']
            except KeyError as error:
                LOGGER.debug('not a CORS request, missing required header %s',
                             error.args[0])
            else:
                self.is_preflight = True
                origin = origin.strip()
                if origin in self.config.allowed_origins:
                    self.requested_origin = origin
                else:
                    LOGGER.debug('CORS preflight origin %r not allowed',
                                 origin)
        else:
            self.is_preflight = False
            origin = request.headers.get('origin', '').strip()
            if origin in self.config.allowed_origins:
                self.requested_origin = origin

        self.set_headers(handler)

    def set_headers(self, handler: web.RequestHandler) -> None:
        """Set the appropriate CORS response headers

        **You probably do not need to invoke this method directly!**

        The response header set is determined by a call to
        :meth:`process_request`.  If it has not been called, then
        only the ``Access-Control-Allow-Origin`` header will be set
        if we are configured to allow any origin.

        """
        if self.is_preflight and self.ok:
            def add_headers(name, values):
                for value in values:
                    handler.add_header(name, value)

            add_headers('Access-Control-Allow-Methods',
                        self.config.allowed_methods)
            add_headers('Access-Control-Allow-Headers',
                        handler.request.headers.get_list(
                            'Access-Control-Request-Headers'))
            add_headers('Access-Control-Expose-Headers',
                        self.config.exposed_headers)
            handler.set_header('Access-Control-Max-Age',
                               str(self.config.max_age))

        if self.requested_origin:
            best_origin = self.requested_origin
            if self.config.allow_credentials:
                handler.set_header('Access-Control-Allow-Credentials', 'true')
            elif self.config.allowed_origins.allow_any:
                best_origin = '*'
            handler.set_header('Access-Control-Allow-Origin', best_origin)
        elif self.config.allowed_origins.allow_any:
            handler.set_header('Access-Control-Allow-Origin', '*')

        handler.add_header('Vary', 'Origin')


class CORSMixin(web.RequestHandler):
    """Mix this in to enable CORS pre-flight and request processing

    This mix-in uses the application level CORS configuration to
    determine how a specific response (or future request) is made
    accessible to a JavaScript client. See the [fetch specification]
    for the gory details.

    [fetch specification]: https://fetch.spec.whatwg.org/#http-cors-protocol

    """

    cors_overrides = {}
    """Add specific overrides for CORSProcessor here"""

    def __init__(self, application, request, **kwargs):
        super().__init__(application, request, **kwargs)
        cors_config = getattr(application, 'cors_config', CORSConfig())
        self.cors = CORSProcessor(cors_config, **self.cors_overrides)

        # This little snippet is responsible for determining which
        # methods are actually implemented and building the
        # "allowed methods" default if it is not configured
        if not self.cors.config.allowed_methods:
            defined_methods = set()
            for method_name in self.SUPPORTED_METHODS:
                method = getattr(self, method_name.lower())
                if method and method.__name__ == method_name.lower():
                    defined_methods.add(method_name)
            defined_methods.discard('OPTIONS')
            self.cors.config.update(allow_methods=defined_methods)

    async def prepare(self):
        maybe_coro = super().prepare()
        if maybe_coro:  # pragma: no cover
            await maybe_coro
        self.cors.process_request(self)

    def options(self) -> None:
        self.set_status(204)

    def set_default_headers(self) -> None:
        super().set_default_headers()
        if hasattr(self, 'cors'):
            self.cors.set_headers(self)
